from ivisit.widgetdefcollection import *
from ivisit.defdb import *
from ivisit.defdialogs import *
from tkinter import *
from PIL import Image, ImageTk
import threading   # for locks
from supy.forms3 import *


# ******************************************
# class IVisitWidget:
# Abstract base class for IVisit Widgets
# ******************************************
class IVisitWidget:
    def __init__(self,parent_arg,widgetdef_arg,className_arg):   
        # set reference parameters
        self.parent = parent_arg        # reference to the parent widget (typically a IVisitRawDisplayFrame)
        self.widgetdef = widgetdef_arg  # reference to the widget definition (either a ParameterWidgetDef or DataWidgetDef)
        self.className=className_arg    # class name for identifying type of widget
        self.sizeX, self.sizeY = 0,0
        self.flagHide=0                 # if set then hide widget
        self.flagModifiedParVal=1       # will be set (1) if parameter value has been modified (only for parameter widgets)
        self.labelFrame=Frame(self)
        self.labelFrame.pack(side=TOP,fill=X)
        self.labeltext=StringVar()
        self.labeltext.set(self.widgetdef.name)
        #self.label=Label(self.labelFrame,text=self.widgetdef.name)
        self.label=Label(self.labelFrame,textvariable=self.labeltext)    # allow modification of text during simulation
        self.label.bind("<Button-1>",self.onLabelPressed)
        self.label.bind("<B1-Motion>",self.onLabelMoved)
        self.label.bind("<ButtonRelease-1>",self.onLabelReleased)
        self.label.bind("<Button-3>",self.onLabelPressedContext)
        self.labelFrame.bind("<Button-1>",self.onLabelPressed)
        self.labelFrame.bind("<B1-Motion>",self.onLabelMoved)
        self.labelFrame.bind("<ButtonRelease-1>",self.onLabelReleased)
        self.labelFrame.bind("<Button-3>",self.onLabelPressedContext)
        self.labelMoveFlag=0
        #self.labelMoveOldX,labelMoveOldY=self.posX,self.posY

    def updateData(self):   # return (width,height) tuple 
        assert 0, "updateData(self) must be implemented in subclasses of IVisitWidget!"

    def setLabelText(self,txt):
        self.labeltext.set(txt)
        
    def onLabelPressed(self,event):
        if(self.labelMoveFlag==0):
            #print "pressed at", event.x,event.y
            self.labelMoveInitX,self.labelMoveInitY = event.x, event.y
            self.labelMoveFlag=1
            self.lift()
        
    def onLabelMoved(self,event):
        if(self.labelMoveFlag>0):
            self.parent.isModified=True
            self.widgetdef.flagModified=True
            #print "moved to", event.x,event.y
            pos=self.widgetdef.pos    # reference to position list
            deltax, deltay = event.x-self.labelMoveInitX, event.y-self.labelMoveInitY    # relative movement of the mouse
            pos[0],pos[1] = pos[0]+deltax, pos[1]+deltay
            if(pos[0]<0): pos[0]=0
            if(pos[1]<0): pos[1]=0
            self.place(x=pos[0], y=pos[1]) 
            #self.update_idletasks()
        
    def onLabelReleased(self,event):
        if(self.labelMoveFlag>0):
            self.parent.adaptSize()
            #print "released at", event.x,event.y
            self.labelMoveFlag=0
            #deltax, deltay = event.x-self.labelMoveInitX, event.y-self.labelMoveInitY    # relative movement of the mouse
            #self.posX, self.posY = self.posX+deltax, self.posY+deltay
            #self.place(x=self.posX, y=self.posY) 
            
    def onLabelPressedContext(self,event):
        w=Toplevel(self.parent)
        if self.parent: w.transient(self.parent)     # make window transient, e.g., minmize with parent etc.
        w.title("Help")
        l=Label(w,text=self.widgetdef.comment)
        l.pack()
        b=Button(w,text="Close",command=w.destroy)
        b.pack()

    def bind(self,eventType,function2call,obj=None):  # bind function2call to object obj if eventType happens
        print("Warning: Call to abstract bind method of IVisitWidget ", self.widgetdef.name)
        return None
        
# *********************************************************
# class IVisitImageWidget:
# IVisit Image Widget to display 2D matrixese as images 
# *********************************************************

class IVisitImageWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitImageWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,DataWidgetDef), "IVisitImageWidget needs to be initialized with a DataWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='image', "IVisitImageWidget needs to be initialized with a DataWidgetDef having type='image', but actually type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.flagRefData>0, "IVisitImageWidget must be initialized with reference data, but self.widgetdef="+str(self.widgetdef.name)+" has no reference data!"
        self.label.pack(side=LEFT,fill=X)
        self.phim=None
        self.imgcanvas=Canvas(self)
        self.myscale=1.0
        self.mychannel="RGB"
        self.setState()
        self.imgcanvas.pack(side=BOTTOM)
        self.menuLabel=Label(self.labelFrame,text="...")
        self.menuLabel.pack(side=LEFT)
        self.menubar=Menu(self.labelFrame,tearoff=0)
        self.channel_menu = Menu(self.menubar,tearoff=0)
        self.channel_menu.add_command(label="RGB",command=lambda:self.setChannel("RGB"))
        self.channel_menu.add_command(label="Gray" ,command=lambda:self.setChannel("Gray"))
        self.channel_menu.add_command(label="Hue as RGB" ,command=lambda:self.setChannel("Hue"))
        self.channel_menu.add_separator()
        self.channel_menu.add_command(label="Red R" ,command=lambda:self.setChannel("R"))
        self.channel_menu.add_command(label="Green G" ,command=lambda:self.setChannel("G"))
        self.channel_menu.add_command(label="Blue B" ,command=lambda:self.setChannel("B"))
        self.channel_menu.add_separator()
        self.channel_menu.add_command(label="Hue H" ,command=lambda:self.setChannel("H"))
        self.channel_menu.add_command(label="Saturation S" ,command=lambda:self.setChannel("S"))
        self.channel_menu.add_command(label="Value V" ,command=lambda:self.setChannel("V"))
        self.menubar.add_cascade(label="Channels",menu=self.channel_menu)
        self.scale_menu = Menu(self.menubar,tearoff=0)
        self.scale_menu.add_command(label="1x",command=lambda:self.setScale(1.0))
        self.scale_menu.add_separator()
        self.scale_menu.add_command(label="2x",command=lambda:self.setScale(2.0))
        self.scale_menu.add_command(label="4x",command=lambda:self.setScale(4.0))
        self.scale_menu.add_command(label="8x",command=lambda:self.setScale(8.0))
        self.scale_menu.add_separator()
        self.scale_menu.add_command(label="1/2x",command=lambda:self.setScale(0.5))
        self.scale_menu.add_command(label="1/4x",command=lambda:self.setScale(0.25))
        self.scale_menu.add_command(label="1/8x",command=lambda:self.setScale(0.125))
        self.menubar.add_cascade(label="Scale",menu=self.scale_menu)
        self.menubar.add_command(label="Properties",command=self.onProperties)
        self.menubar.add_command(label="Save image",command=self.onSaveImage)
        self.menubar.add_command(label="Hide",command=self.onHide)
        #self.b.pack(side=RIGHT)
        self.menuLabel.bind("<Button-1>", self.popup)
        self.menuLabel.bind("<Button-3>", self.popup)
        if(className=="IVisitImageWidget"): self.updateData()      # derived classes may need further preprocessing before call

    def popup(self,event):
        self.menubar.post(event.x_root, event.y_root)

    def setChannel(self,str_channel):
        self.mychannel=str_channel
        self.updateData()
        
    def setScale(self,sc):
        self.myscale=sc
        self.setState()
        self.updateData()

    def onProperties(self):
        self.widgetdef.setDataFromSimulation()
        updated=editSQLTables(self,self.widgetdef.wdefcoll.db,"Editing of Table 'DataWidget'",tbed_datawidget_cfg,cond_on=[["datawidget.key",self.widgetdef.key]])
        if updated:
            self.widgetdef.wdefcoll.initFromDatabase(self.widgetdef.wdefcoll.simulation_key)
            self.parent.master.updateDisplay()
            
    def onSaveImage(self):
        if not self.phim is None:
            ftypes = [('PNG files' , '.png'),
                      ('JPEG files', '.jpg'),
                      ('TIF files' , '.tif'),
                      ('EPS files' , '.eps'),
                      ('All files' , '*'   )]
            fname = tkinter.filedialog.asksaveasfilename(filetypes=ftypes, defaultextension='.png')
            if not fname is None and fname!="":
                imgpil = ImageTk.getimage(self.phim)
                if imgpil.mode in ("RGBA", "P") and fname.split(".")[-1] in ['jpg','JPG','eps','EPS']:
                    imgpil=imgpil.convert("RGB")
                imgpil.save(fname) 
                imgpil.close()    

    def onHide(self):
        self.flagHide=1-self.flagHide

    def setState(self):        # set object fields that are derived from widget and data definitions
        self.shape=np.array(self.widgetdef.data).shape
        assert len(self.shape)>=2, "IVisitImageWidget '"+str(self.widgetdef.name)+"' is of type 'image', but is associatied to a dataarray '" + str(self.widgetdef.name_dataarray) + "' with less than 2 dimensions!" 
        self.RGBflag=0         # default a 2D gray scale array
        if len(self.shape)>=3: self.RGBflag=1    # 3D RGB array?
        self.NY, self.NX = self.shape[0],self.shape[1]
        myscale=self.widgetdef.scale*self.myscale
        self.CX, self.CY = int(myscale*self.NX), int(myscale*self.NY)
        self.SZ          = self.CX * self.CY
        if self.flagHide<=0: self.imgcanvas.configure(width=self.CX,height=self.CY)
        else: self.imgcanvas.configure(width=1,height=1)
        self.sizeX, self.sizeY = max(self.imgcanvas.winfo_reqwidth(),self.label.winfo_reqwidth()), self.imgcanvas.winfo_reqheight()+self.label.winfo_reqheight()


    def updateData(self):
        # (i) convert image?
        if self.RGBflag>0:
            if self.mychannel=="Gray":
                self.im=self.im.convert("L")
                self.label.configure(text=self.widgetdef.name+" (Gray)")
            elif self.mychannel=="R":
                self.im=Image.fromarray(np.array(self.widgetdef.data[:,:,0],dtype=np.uint8))
                self.label.configure(text=self.widgetdef.name+" (R)")
            elif self.mychannel=="G":
                self.im=Image.fromarray(np.array(self.widgetdef.data[:,:,1],dtype=np.uint8))
                self.label.configure(text=self.widgetdef.name+" (G)")
            elif self.mychannel=="B":
                self.im=Image.fromarray(np.array(self.widgetdef.data[:,:,2],dtype=np.uint8))
                self.label.configure(text=self.widgetdef.name+" (B)")
            elif self.mychannel=="H":
                self.im=Image.fromarray(np.array(self.widgetdef.data,dtype=np.uint8))
                self.im=self.im.convert("HSV")
                im=np.array(self.im.getdata(),dtype=np.uint8).reshape(self.widgetdef.data.shape)
                self.im=Image.fromarray(im[:,:,0])
                self.label.configure(text=self.widgetdef.name+" (H)")
            elif self.mychannel=="S":
                self.im=Image.fromarray(np.array(self.widgetdef.data,dtype=np.uint8))
                self.im=self.im.convert("HSV")
                im=np.array(self.im.getdata(),dtype=np.uint8).reshape(self.widgetdef.data.shape)
                self.im=Image.fromarray(im[:,:,1])
                self.label.configure(text=self.widgetdef.name+" (S)")
            elif self.mychannel=="V":
                self.im=Image.fromarray(np.array(self.widgetdef.data,dtype=np.uint8))
                self.im=self.im.convert("HSV")
                im=np.array(self.im.getdata(),dtype=np.uint8).reshape(self.widgetdef.data.shape)
                self.im=Image.fromarray(im[:,:,2])
                self.label.configure(text=self.widgetdef.name+" (V)")
            elif self.mychannel=="Hue":
                self.im=Image.fromarray(np.array(self.widgetdef.data,dtype=np.uint8))
                self.im=self.im.convert("HSV")
                im=np.array(self.im.getdata(),dtype=np.uint8)
                im=im.reshape(self.widgetdef.data.shape)
                print("im.shape=",im.shape)
                im[:,:,1]=255
                im[:,:,2]=255
                self.im=Image.fromarray(im,mode='HSV')
                self.label.configure(text=self.widgetdef.name+" (HSV-Hue)")
            else:
                self.im=Image.fromarray(np.array(self.widgetdef.data,dtype=np.uint8))  # default
                self.label.configure(text=self.widgetdef.name)
        else:
            self.im=Image.fromarray(np.array(self.widgetdef.data,dtype=np.uint8))      # default
            self.label.configure(text=self.widgetdef.name)
        # (ii) rescale image size?
        myscale=self.widgetdef.scale*self.myscale
        if(myscale!=1.0):
            if(myscale<1.0): self.im=self.im.resize((self.CX,self.CY),Image.ANTIALIAS)
            else: self.im=self.im.resize((self.CX,self.CY),Image.NEAREST)
        # (iii) display image
        self.phim = ImageTk.PhotoImage(self.im)
        self.imgcanvas.delete('all');
        if self.flagHide<=0: self.imgcanvas.create_image(0,0,image=self.phim,anchor=NW)
        self.setState()
        #self.imgcanvas.create_line(0,0,100,100)

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.CX,self.CY)

    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the image canvas (e.g., if clicking on an image pixel)
        if obj==None:
            obj="imgcanvas"
        if obj=="imgcanvas":
            self.imgcanvas.bind(eventType,function2call)
            return self.imgcanvas
        else:
            print("Warning in IVisitImageWidget.bind(): Unknown widget object ", obj, " in IVisitImageWidget", self.widgetdef.name)
            return None
        
# *********************************************************
# class IVisitTextfieldWidget:
# IVisit Image Widget to display text data  
# *********************************************************

class IVisitTextfieldWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitTextfieldWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,DataWidgetDef), "IVisitTextfieldWidget needs to be initialized with a DataWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='textfield', "IVisitTextfieldWidget needs to be initialized with a DataWidgetDef having type='textfield', but actually type="+str(self.widgetdef.type)+" is not valid!" 
        self.label.pack(side=LEFT)
        options=self.widgetdef.range
        str_just="center"
        if 'just_left'  in options: str_just="left"
        if 'just_right' in options: str_just="right"
        self.textlabel=Label(self,text="",justify=str_just)
        self.textlabel.pack(side=RIGHT)
        self.setState()
        if(className=="IVisitTextfieldWidget"): self.updateData()      # derived classes may need further preprocessing before call

    def setState(self):        # set object fields that are derived from widget and data definitions
        self.sizeX, self.sizeY = max(self.label.winfo_reqwidth()+self.textlabel.winfo_reqwidth()+10,self.winfo_reqwidth()), max(self.winfo_reqheight(),self.label.winfo_reqheight(),self.textlabel.winfo_reqheight())

    def updateData(self):
        if self.widgetdef.scale>0:
            self.label.configure(text=str(self.widgetdef.name)+': ')
        else:
            self.label.configure(text="")
        self.textlabel.configure(text=str(self.widgetdef.data))
        self.setState()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
        
'''
# *******************************************************************
# class PyFeCombiImageWidget:
# PyFelix CombiImage Widget to display 2D population variables plus
# one binary 2D population variable,
# for example, spikes and dendritic potentials
# programmed on September 21, 2012 by Andreas Knoblauch
# *******************************************************************

# color palette for PyFeCombiImageWidget
PyFeCombiImageColorPalette = [item for sublist in ([[i,i,i] for i in range(255)] + [[255,0,0]]) for item in sublist]   

class PyFeCombiImageWidget(PyFeImageWidget):
    def __init__(self,parent,widgetdef_arg,database_arg):
        Frame.__init__(self,master=parent)
        PyFeImageWidget.__init__(self,parent,widgetdef_arg,database_arg,"PyFeCombiImageWidget")
        self.im.putpalette(PyFeCombiImageColorPalette)
        self.updateData()

    def updateData(self):
        #print "call to PyFeInter_getGradualCombiImage"
        self.database.libSim.PyFeInter_getGradualCombiImage(self.data_idx,self.scaleFactor,c_float(self.rangeMin),c_float(self.rangeMax),\
                                                            254,byref(self.imarray))
        self.im.fromstring(self.imarray)
        self.phim = ImageTk.PhotoImage(self.im)
        self.imgcanvas.delete('all');
        self.imgcanvas.create_image(0,0,image=self.phim,anchor=NW)
        #self.imgcanvas.create_line(0,0,100,100)
'''
       
        
# *********************************************************
# class IVisitSliderWidget:
# IVisit Slider Widget to control a simulation parameter
# *********************************************************

class IVisitSliderWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitSliderWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitSliderWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='slider', "IVisitSliderWidget needs to be initialized with a ParameterWidgetDef having type='slider', but actually type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.type_parameter=='int' or self.widgetdef.type_parameter=='float', "IVisitSliderWidget needs to be initialized with a numeric parameter type, but actually parameter_type="+str(self.widgetdef.type_parameter)+" is not valid!"
        assert isinstance(self.widgetdef.range,(tuple,list)) and len(self.widgetdef.range)>=4, "IVisitSliderWidget '"+str(self.widgetdef.name)+"' must have attribute range=[min,max,nTicks,scale], but actually range="+str(self.widgetdef.range)+"!"
        self.label.pack(side=LEFT)
        assert(self.widgetdef.range[0]<=self.widgetdef.range[1])and(self.widgetdef.range[3]>=0),"IVisitSliderWidget.__init__: wrong slider range parameters for slider widget " +\
            self.widgetdef.name + "! It must be min<=max and scale>0, whereas range=" + str(self.widgetdef.range) + "! Have a look at range specification in the corresponding " +\
            "parameter definition in the ParameterWidget table! Check format range=[min,max,nTicks,scale] after parameter name!"
        min_,max_,nTicks,scale=self.widgetdef.range[0],self.widgetdef.range[1],int(np.round(self.widgetdef.range[2])),self.widgetdef.range[3]
        if(self.widgetdef.type_parameter=='int'):
            min_,max_,scale = int(np.round(min_)),int(np.round(max_)),int(np.round(scale))
            #print("widget=",self.widgetdef.name,"min,max,scale=",min_,max_,scale)
        # ACHTUNG! BUG IN TKINTER: Man kann Slider nicht für ungerade integer-Werte konfigurieren (nur für gerade)!!!!!!!!!! Stand: Python 3.6 !!!! Vielleicht behoben mit höheren Versionen???
        self.slider=Scale(self,from_=min_,to=max_,command=self.onMove,orient='horizontal',\
                          tickinterval=(max_-min_)/nTicks,resolution=scale,length=int(self.widgetdef.size[0]))
        self.slider.set(self.widgetdef.value)
        self.slider.pack(side=LEFT)
        self.callback_on_sliderupdate=None
        self.onMove(self.widgetdef.value)
        self.setState()

    def setState(self):        # set object fields that are derived from widget and data definitions
        #min=c_double()    # ?????
        self.sizeX, self.sizeY = self.slider.winfo_reqwidth()+self.label.winfo_reqwidth(), self.slider.winfo_reqheight()+self.label.winfo_reqheight()
        #print "slider.sizeX=", self.sizeX, " sizeY=", self.sizeY

    def setValue(self,val):
        self.slider.set(val)

    def reconfigure(self,min=None,max=None,nTicks=None,scale=None):
        if min is None: min=self.widgetdef.range[0]
        if max is None: max=self.widgetdef.range[1]
        if nTicks is None: nTicks=int(np.round(self.widgetdef.range[2]))
        if scale is None: scale=self.widgetdef.range[3]
        if(self.widgetdef.type_parameter=='int'):
            min,max,scale = int(np.round(min)),int(np.round(max)),int(np.round(scale))
        self.slider.configure(from_=min,to=max,tickinterval=(max-min)/nTicks,resolution=scale)
        
    def onMove(self, val):
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        newvalue = float(val)     #/self.data.scale
        if self.widgetdef.type_parameter=='int':
           newvalue = int(round(newvalue))
        if(newvalue!=self.widgetdef.value):
            self.parent.isModified=True
            self.flagModifiedParVal=1
        #print("value,newvalue=",self.widgetdef.value, newvalue)
        self.widgetdef.value=newvalue  #  or int(float(val)/self.data.scale+0.5)) ????
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()     # write new parameter value through to simulation (only for updateMode 'async')
        #print "new slider value is ", self.database.libSim.PyFeInter_getSliderValue(self.data_idx)*self.data.scale
        if not self.callback_on_sliderupdate is None: self.callback_on_sliderupdate()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the slider (e.g., if selecting a new value)
        if obj==None:
            obj="slider"
        if obj=="slider":
            self.callback_on_sliderupdate=function2call
        else:
            print("Warning in IVisitSliderWidget.bind(): Unknown widget object ", obj, " in IVisitSliderWidget", self.widgetdef.name)
        return None
            
# **************************************************************************************
# class IVisitDictSliderWidget:
# IVisit DictSlider Widget to control a dict of several different simulation parameters
# **************************************************************************************

class IVisitDictSliderWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitDictSliderWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        self.widgetdef.ref_to_dictsliderwidget=self   # DIRTY!!! ALLOW widgetdef object to efficiently access values !!!!!!!!! See widgetdefcollection.py; ParameterWidgetDef.setSimParValue(self)
        # (i) check consistency of widgetdef attributes and extract relevant information
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitDictSliderWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='dictslider', "IVisitDictSliderWidget needs to be initialized with a ParameterWidgetDef having type='dictslider', but actually type="+str(self.widgetdef.type)+" is not valid!"
        wdef=self.widgetdef
        msg_format="IVisitDictSliderWidget: widgetdef.items must have format: [nitems,idxselecteditem,  itemname1,type1,  itemname2,typ2, ...] (length 2+nitems*2)\n"
        msg_format+=" widgetdef.range must have format: [nitems, itemname1,min1,max1,nticks1,scale1, itemname2,min2,max2,nticks2,scale2, ...] (length 1+nitems*5\n"
        msg_format+=" widgetdef.range_parameter must have format: [nitems, key1,type1,min1,max1, key2,type2,min2,max2, ...] (length 1+nitems*4\n"
        assert isinstance(wdef.items,list) and len(wdef.items)>0,"IVisitDictSliderWidget "+str(self.widgetdef.name)+" must have wdef.items being a non-empty list, but wdef.items="+str(wdef.items)+"\n"+msg_format
        assert isinstance(wdef.range,list) and len(wdef.range)>0,"IVisitDictSliderWidget "+str(self.widgetdef.name)+" must have self.parameter being a non-empty list, but self.range="+str(self.range)+"\n"+msg_format
        assert isinstance(wdef.range_parameter,list) and len(wdef.range_parameter)>0,"IVisitDictSliderWidget "+str(self.widgetdef.name)\
            +" must have self.range_parameter being a non-empty list, but self.range_parameter="+str(self.range_parameter)+"\n"+msg_format
        self.nitems=wdef.items[0]         # for format see above
        assert isinstance(self.nitems,int),"wdef.items[0] must be int, but wdef.items="+wdef.items+"\n"+str(msg_format)
        assert len(wdef.items)==2+self.nitems*2,"wdef.items has wrong length!\n"+msg_format+"but wdef.items="+str(wdef.items)
        assert len(wdef.range)==1+self.nitems*5,"wdef.range has wrong length!\n"+msg_format+"but wdef.range="+str(wdef.range) 
        assert len(wdef.range_parameter)==1+self.nitems*4,"wdef.range_parameter has wrong length!\n"+msg_format+"but wdef.range_parameter="+str(wdef.range_parameter)
        self.itemtypes=[wdef.items[3+i*2] for i in range(self.nitems)]   # types of items
        for i in range(self.nitems):
            assert self.itemtypes[i] in ['int','float'],"All item types must be either 'int' or 'float', but self.itemtypes="+str(self.itemtypes)
            ibase=1+i*5
            assert isinstance(wdef.range[ibase],str) and isinstance(wdef.range[ibase+1],(int,float)) and isinstance(wdef.range[ibase+2],(int,float)) \
                and isinstance(wdef.range[ibase+3],int) and isinstance(wdef.range[ibase+4],(int,float)),"wdef.range has wrong types! But wdef.range="+str(wdef_range)+"\n"+msg_format
            assert wdef.range[ibase+1]<wdef.range[ibase+2] and wdef.range[ibase+4]!=0,"wdef.range has wrong mini,maxi,scalei! But wdef.range="+str(wdef_range)+"\n"+msg_format+"\n"+msg_format
            ibasep=1+i*4
            assert wdef.range[ibase+1]<wdef.range[ibase+2],"wdef.range.min < max required!\nBut wdef.range="+str(wdef.range)+"\n"+msg_format
            assert wdef.range_parameter[ibasep+2]<wdef.range_parameter[ibasep+3],"wdef.range_parameter.min < max required!\nBut wdef.range_parameter="+str(wdef.range_parameter)+"\n"+msg_format
        self.sliderranges=[[wdef.range[i*5+2],wdef.range[i*5+3],wdef.range[i*5+4],wdef.range[i*5+5]] for i in range(self.nitems)] # range information for each item for slider widget
        self.itemnames=[wdef.range[1+i*5] for i in range(self.nitems)]                                                            # item names for dict keys (to be displayed in item list of widget)
        self.dict_itemnames={self.itemnames[i]:i for i in range(self.nitems)}                                                     # dict for association itemname --> index
        self.itemkeys=[wdef.range_parameter[1+i*4] for i in range(self.nitems)]                                                   # item keys of dict
        self.values=parseStringAsList(wdef.value,'string',[])                                                                     # current values: first get list of string values
        for i in range(self.nitems):                                                                                              # ...then cast string values as numerical values
            if self.itemtypes[i]=='int': self.values[i]=asNumber(self.values[i],'int',0)                                          # ...either as int
            else: self.values[i]=asNumber(self.values[i],'float',0.0)                                                             # ...or float
        assert isinstance(self.values,list) and len(self.values)==self.nitems,"wdef.value must be string of list of length nitems="+str(self.nitems)+", but wdef.value="+str(wdef.value)
        self.selecteditem=wdef.items[1]                                                                                           # currently selected item
        self.selectedvalue=self.values[self.selecteditem]                                                                         # value of currently selected item
        sz=wdef.size
        self.size_sld,self.columns_text,self.rows_text,self.disp_mode,self.fontsz = 200,20,-1,0,10   # default values for slider size, number of columns in text widget, number of rows, display mode and fontsize
        if isinstance(sz,list):
            if len(sz)>0: self.size_sld=int(sz[0])            # slider size (=width)
            if len(sz)>1: self.columns_text=int(sz[1])        # number of text columns of text widget (for disp_mode=2) 
            if len(sz)>2: self.rows_text=int(sz[2])           # number of text rows of text widget (for disp_mode=2) 
            if len(sz)>3: self.disp_mode=int(sz[3])           # display mode (0=no extra; 1=display parameter values in optionlist entries; 2=extra textfield for parameter values)
            if len(sz)>4: self.fontsz=int(sz[4])              # fontsize for parameter values in textfield (only for disp_mode=2)
        else:
            self.size_sld=int(sz)
        self.itemnames4optionmenu=[s for s in self.itemnames] # default: use simply itemnames for optionmenu
        if self.disp_mode==1:
            self.itemnames4optionmenu=[self.itemnames[i]+": "+str(self.values[i]) for i in range(self.nitems)]   # for dips_mode1 display parameter values in option menu
        #print("IVisitDictSliderWidget",wdef.name)
        #print("   nitems=",self.nitems)
        #print("   itemnames=",self.itemnames)
        #print("   itemkeys=",self.itemkeys)
        #print("   itemtypes=",self.itemtypes)
        #print("   sliderranges=",self.sliderranges)
        #print("   selecteditem=",self.selecteditem)
        #print("   values=",self.values)
        #print("   selectedvalue=",self.selectedvalue)
        #exit(0)
        # (ii) create and initialize widgets
        # (ii.1) do some preparations
        self.label.pack(side=LEFT)
        self.parameterFrame=Frame(self)
        self.parameterFrame.pack(side=BOTTOM,expand=True,fill=BOTH)
        # (ii.2) option menu
        self.itemname_str_var = StringVar()
        self.itemname_str_var.set(str(self.itemnames4optionmenu[self.selecteditem]))
        self.optionmenu = OptionMenu(self.labelFrame,self.itemname_str_var,command=self.onSet,*self.itemnames4optionmenu)
        self.optionmenu.pack(side=LEFT)
        # (ii.3) slider
        rng=self.sliderranges[self.selecteditem]
        min,max,nTicks,scale=rng[0],rng[1],int(np.round(rng[2])),rng[3]
        if(self.itemtypes[self.selecteditem]=='int'):
            min,max,scale = int(np.round(min)),int(np.round(max)),int(np.round(scale))
        self.slider=Scale(self.parameterFrame,from_=min,to=max,command=self.onMove,orient='horizontal',\
                          tickinterval=(max-min)/nTicks,resolution=scale,length=int(self.size_sld))
        self.slider.set(self.values[self.selecteditem])
        self.slider.pack(side=TOP)
        # (ii.4) text widget
        if self.disp_mode==2:
            self.textFrame=Frame(self.parameterFrame)
            self.textFrame.pack(side=BOTTOM,expand=True,fill=BOTH)
            self.text_str_var = StringVar()
            self.text_str_var.set("".join([self.itemnames[i]+": "+str(self.values[i])+"\n" for i in range(self.nitems)])[:-1])
            if self.rows_text<=0: self.rows_text=self.nitems
            self.textinput = SupyTextFrame(self.textFrame, width=self.columns_text,height=self.rows_text, catch_focus=False, textvariable=self.text_str_var)
            self.textinput.text.configure(font=('Arial',self.fontsz))
            self.textinput.pack(side=LEFT)
            self.setbutton=Button(self.textFrame,text='Set',command=self.onSetTextButton)
            self.setbutton.pack(side=RIGHT)

        # (iii) finalize
        self.callback_on_sliderupdate=None
        self.callback_on_optionmenuupdate=None
        self.callback_on_textupdate=None
        self.onMove(self.values[self.selecteditem])
        self.setState()
        self.reconfigureOptionMenu()

    def setState(self):        # set object fields that are derived from widget and data definitions
        #min=c_double()    # ?????
        self.sizeX = max(self.slider.winfo_reqwidth(),self.label.winfo_reqwidth()+self.optionmenu.winfo_reqwidth())
        self.sizeY = max(self.label.winfo_reqheight(),self.optionmenu.winfo_reqheight())+self.slider.winfo_reqheight()
        #print "slider.sizeX=", self.sizeX, " sizeY=", self.sizeY

    def setValue(self,val):
        self.values=[val[i] for i in range(self.nitems)]
        self.selectedvalue=self.values[self.selecteditem]
        self.slider.set(self.selectedvalue)
        #sliderget=self.slider.get()
        #if self.selectedvalue!=sliderget:           # has slider modified value (due to range restrictions)? NO!!!!! THEN let the value be!!! Maybe useful to test parameter values outside the slider range/grid !!??
        #    if self.itemtypes[self.selecteditem]=='int': self.selectedvalue=int(sliderget)
        #    else                                       : self.selectedvalue=float(sliderget)
        #self.values[self.selecteditem]=self.selectedvalue
        #print("...final selectedalue=",self.selectedvalue)
        self.reconfigureOptionMenu()

    def reconfigureSlider(self,min=None,max=None,nTicks=None,scale=None):  
        rng=self.sliderranges[self.selecteditem]
        min,max,nTicks,scale=rng[0],rng[1],int(np.round(rng[2])),rng[3]
        if min is None: min=rng[0]
        if max is None: max=rng[1]
        if nTicks is None: nTicks=int(np.round(rng[2]))
        if scale is None: scale=rng[3]
        if(self.itemtypes[self.selecteditem]=='int'):
            min,max,scale = int(np.round(min)),int(np.round(max)),int(np.round(scale))
        self.slider.configure(from_=min,to=max,tickinterval=(max-min)/nTicks,resolution=scale)

    def reconfigureOptionMenu(self,item_idx=None):
        if item_idx is None:
            item_idx=list(range(self.nitems))
        elif isinstance(item_idx,int):
            item_idx=[item_idx]
        if self.disp_mode==1:
            for i in item_idx:
                self.itemnames4optionmenu[i]=self.itemnames[i]+": "+str(self.values[i])
        else:
            for i in item_idx:
                self.itemnames4optionmenu[i]=self.itemnames[i]
        for i in item_idx: self.optionmenu['menu'].entryconfigure(i, label=self.itemnames4optionmenu[i])
        if self.selecteditem in item_idx: self.itemname_str_var.set(self.itemnames4optionmenu[self.selecteditem])
        
    def onSet(self,e=None):                                      # called by optionmenu if new item is selected
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        newitemname = self.itemname_str_var.get().split(':')[0]  # for disp_mode=1 take only part left of (last) ':'
        newitem = self.dict_itemnames[newitemname]               
        if newitem!=self.selecteditem:
            self.parent.isModified=True
            self.selecteditem=newitem
            self.selectedvalue=self.values[newitem]
            self.reconfigureSlider()
            self.slider.set(self.selectedvalue)
        if not self.callback_on_optionmenuupdate is None: self.callback_on_optionmenuupdate()

    def onMove(self, val):                                # called by slider if new value is selected
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        newvalue = float(val)     # self.data.scale
        if self.itemtypes[self.selecteditem]=='int':
           newvalue = int(round(newvalue))
        #print("dictslider new value=",newvalue, "selectedvalue=",self.selectedvalue)
        if(newvalue!=self.selectedvalue):
            self.parent.isModified=True
            self.flagModifiedParVal=1
        #print("value,newvalue=",self.widgetdef.value, newvalue)
        self.values[self.selecteditem]=newvalue
        self.selectedvalue=newvalue
        self.widgetdef.value=str(self.values)    #  or int(float(val)/self.data.scale+0.5)) ????
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()     # write new parameter value through to simulation (only for updateMode 'async')
        #print "new slider value is ", self.database.libSim.PyFeInter_getDictSliderValue(self.data_idx)*self.data.scale
        if self.disp_mode==1: self.reconfigureOptionMenu(self.selecteditem)   # update parameter value in option menu (only for disp_mode=1)
        elif self.disp_mode==2: self.textinput.settext("".join([self.itemnames[i]+": "+str(self.values[i])+"\n" for i in range(self.nitems)])[:-1])
        if not self.callback_on_sliderupdate is None: self.callback_on_sliderupdate()

    def onSetTextButton(self):
        txt=self.textinput.gettext()
        txt=txt.split('\n')
        listvalstr=[s.split(':')[-1] for s in txt]
        newvalues=self.values
        for i in range(self.nitems):
            if i<len(listvalstr):
                v=asNumber(listvalstr[i],self.itemtypes[i],None)
                if not v is None: newvalues[i]=v
        self.setValue(newvalues)
        self.parent.isModified=True
        self.flagModifiedParVal=1
        self.widgetdef.value=str(self.values)    
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()     # write new parameter value through to simulation (only for updateMode 'async')
        if self.disp_mode==1: self.reconfigureOptionMenu(self.selecteditem)   # update parameter value in option menu (only for disp_mode=1)
        self.textinput.settext("".join([self.itemnames[i]+": "+str(self.values[i])+"\n" for i in range(self.nitems)])[:-1])   # normalize text in textbox
        if not self.callback_on_textupdate is None: self.callback_on_textupdate()
        
            
    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the slider (e.g., if selecting a new value)
        if obj==None:
            obj="slider"
        if obj=="slider":
            self.callback_on_sliderupdate=function2call
        if obj=='optionmenu':
            self.callback_on_optionmenuupdate=function2call
        if obj=='textbox':
            self.callback_on_textupdate=function2call
        else:
            print("Warning in IVisitDictSliderWidget.bind(): Unknown widget object ", obj, " in IVisitDictSliderWidget", self.widgetdef.name)
        return None
            
# ***********************************************************************
# class IVisitTextInputWidget:
# IVisit Text-Input Widget to control a string-type simulation parameter
# ***********************************************************************

class IVisitTextInputWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitTextInputWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitTextInputWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='textfield', "IVisitTextInputWidget needs to be initialized with a ParameterWidgetDef having type='textfield', but actually type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.type_parameter=='text' or self.widgetdef.type_parameter=='string', "IVisitTextInputWidget needs to be initialized with a parameter of type 'text' or 'string', but actually parameter_type="+str(self.widgetdef.type_parameter)+" is not valid!"
        #assert isinstance(self.widgetdef.range,(tuple,list)) and len(self.widgetdef.range)>=4, "IVisitTextInputWidget '"+str(self.widgetdef.name)+"' must have attribute range=[min,max,nTicks,scale], but actually range="+str(self.widgetdef.range)+"!"
        self.label.pack(side=LEFT)
        #assert(self.widgetdef.range[0]<=self.widgetdef.range[1])and(self.widgetdef.range[3]>=0),"IVisitTextInputWidget.__init__: wrong slider range parameters for slider widget " +\
        #    self.widgetdef.name + "! It must be min<=max and scale>0, whereas range=" + str(self.widgetdef.range) + "! Have a look at range specification in the corresponding " +\
        #    "parameter definition in the ParameterWidget table! Check format range=[min,max,nTicks,scale] after parameter name!"
        self.cols,self.rows = self.widgetdef.size[0], self.widgetdef.size[1]
        self.str_var = StringVar()
        self.str_var.set(str(self.widgetdef.value))
        self.textframe = SupyTextFrame(self, width=self.cols, height=self.rows, catch_focus=False, textvariable=self.str_var)
        self.textframe.pack(side=LEFT)
        self.setbutton=Button(self,text='Set',command=self.onSet)
        self.setbutton.pack(side=LEFT)
        self.callback_on_textUpdate = None
        self.onSet() 
        self.setState()

    def set(self,text):     # set text
        self.textframe.settext(str(text))
        print("str_var=",str(text))
        self.onSet()
    
    def setState(self):        # set object fields that are derived from widget and data definitions
        #min=c_double()    # ?????
        self.sizeX, self.sizeY = self.label.winfo_reqwidth()+self.textframe.winfo_reqwidth()+self.setbutton.winfo_reqwidth(),self.label.winfo_reqheight()+self.textframe.winfo_reqheight()+self.setbutton.winfo_reqheight()
        #print "slider.sizeX=", self.sizeX, " sizeY=", self.sizeY

    def onSet(self):
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        newvalue = self.textframe.gettext()
        if(newvalue!=self.widgetdef.value):
            self.parent.isModified=True
            self.flagModifiedParVal=1
        self.widgetdef.value=newvalue  
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()     # write new parameter value to simulation (only for updateMode 'async')
        if not self.callback_on_textUpdate is None: self.callback_on_textUpdate()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)

    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the slider (e.g., if selecting a new value)
        if obj==None:
            obj="textfield"
        if obj=="textfield":
            self.callback_on_textUpdate=function2call
        else:
            print("Warning in IVisitTextInputWidget.bind(): Unknown widget object ", obj, " in IVisitTextInputWidget", self.widgetdef.name)
        return None
        
# ***********************************************************************
# class IVisitListSelectionWidget:
# IVisit Option-List to control a string-type simulation parameter
# ***********************************************************************

class IVisitListSelectionWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitListSelectionWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitTextInputWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='listselection', "IVisitTextInputWidget needs to be initialized with a ParameterWidgetDef having type='listsel', but actual type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.type_parameter in ['text','string','int','float'], "IVisitListSelectionWidget needs to be initialized with a parameter of type 'text' or 'string' or 'int' or 'float', but actual parameter_type="+str(self.widgetdef.type_parameter)+" is not valid!"
        #assert isinstance(self.widgetdef.range,(tuple,list)) and len(self.widgetdef.range)>=4, "IVisitTextInputWidget '"+str(self.widgetdef.name)+"' must have attribute range=[min,max,nTicks,scale], but actually range="+str(self.widgetdef.range)+"!"
        self.label.pack(side=LEFT)
        #assert(self.widgetdef.range[0]<=self.widgetdef.range[1])and(self.widgetdef.range[3]>=0),"IVisitTextInputWidget.__init__: wrong slider range parameters for slider widget " +\
        #    self.widgetdef.name + "! It must be min<=max and scale>0, whereas range=" + str(self.widgetdef.range) + "! Have a look at range specification in the corresponding " +\
        #    "parameter definition in the ParameterWidget table! Check format range=[min,max,nTicks,scale] after parameter name!"
        self.listvalues = self.widgetdef.range    # list of strings to be diplayed in the OptionMenu
        self.listvalues = [str(s) for s in self.listvalues]
        self.str_var = StringVar()
        self.str_var.set(str(self.widgetdef.value))
        #print("ListSel: widgetdef.value=",self.widgetdef.value)
        self.optionmenu = OptionMenu(self,self.str_var,command=self.onSet,*self.listvalues)
        self.optionmenu.pack(side=LEFT)
        self.callback_on_optionmenuupdate=None
        self.setState()
        self.onSet()

    def setState(self):        # set object fields that are derived from widget and data definitions
        #min=c_double()    # ?????
        self.sizeX, self.sizeY = self.label.winfo_reqwidth()+self.optionmenu.winfo_reqwidth(), self.label.winfo_reqheight()+self.optionmenu.winfo_reqheight()
        #print "slider.sizeX=", self.sizeX, " sizeY=", self.sizeY

    def onSet(self,e=None):
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        newvalue = self.str_var.get() 
        if(newvalue!=self.widgetdef.value):
            self.parent.isModified=True
            self.flagModifiedParVal=1
        self.widgetdef.value=newvalue  
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()     # write new parameter value to simulation (only for updateMode 'async')
        if not self.callback_on_optionmenuupdate is None: self.callback_on_optionmenuupdate()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the OptionMenu (e.g., if selecting a new value)
        if obj==None:
            obj="optionmenu"
        if obj=="optionmenu":
            self.callback_on_optionmenuupdate=function2call
        else:
            print("Warning in IVisitListSelectionWidget.bind(): Unknown widget object ", obj, " in IVisitListSelectionWidget", self.widgetdef.name)
        return None
            
# **************************************************************************
# class IVisitCheckboxWidget:
# IVisit Checkbox to control a string of boolean-type simulation parameters
# **************************************************************************

class IVisitCheckboxWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitCheckboxWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitCheckboxWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='checkbox', "IVisitCheckboxWidget "+str(self.widgetdef.name)+" needs to be initialized with a ParameterWidgetDef having type='checkbox', but actual type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.type_parameter in ['string','str','text'], "IVisitCheckboxWidget "+str(self.widgetdef.name)+" needs to be initialized with parameters of type 'text', but actual parameter_type="+str(self.widgetdef.type_parameter)+" is not valid!"
        assert isinstance(self.widgetdef.range,(tuple,list)), "IVisitCheckboxWidget "+str(self.widgetdef.name)+": IVisitCheckboxWidget.widgetdef.range="+str(self.widgetdef.range)+" must be a string list!"
        assert isinstance(self.widgetdef.value,str), "IVisitCheckboxWidget "+str(self.widgetdef.name)+": IVisitCheckboxWidget.widgetdef.value="+str(self.widgetdef.value)+" must be a string!"
        assert len(self.widgetdef.range)==len(self.widgetdef.value), "IVisitCheckboxWidget "+str(self.widgetdef.name)+": IVisitCheckboxWidget.widgetdef.value and range must have same length!"
        self.label.pack(side=LEFT)
        self.checkboxFrame=Frame(self)
        self.checkboxFrame.pack(side=BOTTOM,expand=True,fill=BOTH)
        self.listkeys = self.widgetdef.range    # list of strings to be diplayed in the Checkbox Frame
        self.listvalues = [str(s) for s in self.listkeys]
        self.int_vars = [IntVar() for s in self.listvalues]
        self.checkbuttons=[Checkbutton(self.checkboxFrame, text=self.listkeys[i], variable=self.int_vars[i], command=self.onSet, justify=LEFT, anchor="w") for i in range(len(self.int_vars))]
        for i in range(len(self.int_vars)):
            assert self.widgetdef.value[i] in ['0','1'], "IVisitCheckboxWidget "+str(self.widgetdef.name)+": IVisitCheckboxWidget.widgetdef.value="+str(self.widgetdef.value)+" must be a string of '0' and '1'!"
            self.int_vars[i].set(int(self.widgetdef.value[i]))
            self.checkbuttons[i].grid(row=i,sticky=W)
        self.callback_on_checkboxupdate=None
        self.setState()
        self.onSet()

    def setState(self):        # set object fields that are derived from widget and data definitions
        self.sizeX = max([self.winfo_reqwidth()]+[self.label.winfo_reqwidth()]+[cb.winfo_reqwidth() for cb in self.checkbuttons])
        self.sizeY = self.label.winfo_reqheight()+np.sum([cb.winfo_reqheight() for cb in self.checkbuttons])

    def setValue(self,val):
        assert isinstance(val,str), "IVisitCheckboxWidget "+str(self.widgetdef.name)+": parameter val="+str(val)+" must be string of '0' and '1'"
        assert len(val)==len(self.int_vars), "IVisitCheckboxWidget "+str(self.widgetdef.name)+": length len(val)="+str(len(val))+" must be the same as number of checkboxes "+str(len(self.int_vars))
        for i in range(len(val)):
            assert val[i] in ['0','1'], "parameter val="+str(val)+" must be string of '0' and '1'"
            self.int_vars[i].set(int(val[i]))

    def onSet(self,e=None):
        newvalue = "".join([str(iv.get()) for iv in self.int_vars])
        if newvalue!=self.widgetdef.value:
            self.parent.isModified=True  # newvalue differs from previous value?
            self.flagModifiedParVal=1
        self.widgetdef.value=newvalue                                   # copy newvalue 
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()           # write new parameter value to simulation (only for updateMode 'async')
        if not self.callback_on_checkboxupdate is None: self.callback_on_checkboxupdate()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the OptionMenu (e.g., if selecting a new value)
        if obj==None:
            obj="checkbox"
        if obj=="checkbox":
            self.callback_on_checkboxupdate=function2call
        else:
            print("Warning in IVisitCheckboxWidget.bind(): Unknown widget object ", obj, " in IVisitCheckboxWidget", self.widgetdef.name)
        return None
            
# ***********************************************************************
# class IVisitRadiobuttonWidget:
# IVisit Radiobutton to control a string of boolean-type simulation parameters
# ***********************************************************************

class IVisitRadiobuttonWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitRadioButtonWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitRadiobuttonWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='radiobutton', "IVisitRadiobuttonWidget needs to be initialized with a ParameterWidgetDef having type='radiobutton', but actual type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.type_parameter in ['string','str','text'], "IVisitRadiobuttonWidget needs to be initialized with parameters of type 'text', but actual parameter_type="+str(self.widgetdef.type_parameter)+" is not valid!"
        assert isinstance(self.widgetdef.range,(tuple,list)), "IVisitRadiobuttonWidget.widgetdef.range="+str(self.widgetdef.range)+" must be a string list!"
        assert isinstance(self.widgetdef.value,str), "IVisitRadiobuttonWidget.widgetdef.value="+str(self.widgetdef.value)+" must be a string!"
        self.label.pack(side=LEFT)
        self.radiobuttonFrame=Frame(self)
        self.radiobuttonFrame.pack(side=BOTTOM,expand=True,fill=BOTH)
        self.listkeys = [str(s) for s in self.widgetdef.range]  # list of strings to be diplayed in the Radiobutton Frame
        self.str_var = StringVar()
        self.radiobuttons=[Radiobutton(self.radiobuttonFrame, text=self.listkeys[i], variable=self.str_var, command=self.onSet, justify=LEFT, anchor="w", value=self.listkeys[i]) for i in range(len(self.listkeys))]
        self.dict_radiobuttons={self.listkeys[i]:self.radiobuttons[i] for i in range(len(self.listkeys))}
        assert str(self.widgetdef.value) in self.listkeys, \
            "IVisitRadiobuttonWidget needs to be initialized a value in listkeys="+str(self.listkeys)+" but widgetdef.value="+str(self.widgetdef.value)+" is not contained in this list!"
        self.str_var.set(str(self.widgetdef.value))
        for i in range(len(self.radiobuttons)):
            self.radiobuttons[i].grid(row=i,sticky=W)
        self.callback_on_radiobuttonupdate=None
        self.setState()
        self.onSet()

    def setState(self):        # set object fields that are derived from widget and data definitions
        self.sizeX = max([self.winfo_reqwidth()]+[self.label.winfo_reqwidth()]+[rb.winfo_reqwidth() for rb in self.radiobuttons])
        self.sizeY = self.label.winfo_reqheight()+np.sum([rb.winfo_reqheight() for rb in self.radiobuttons])

    def setValue(self,val,flagInvoke=0):
        assert isinstance(val,str), "parameter val="+str(val)+" must be a string!"
        assert val in self.listkeys, "string parameter val="+str(val)+" must be from key list "+str(self.listkeys)
        self.str_var.set(val)
        if flagInvoke>0:
            self.dict_radiobuttons[val].invoke()   # is as the user has pressed radio button (call to onSet etc.)

    def onSet(self,e=None):
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        newvalue = str(self.str_var.get())
        if newvalue!=self.widgetdef.value:
            self.parent.isModified=True  # newvalue differs from previous value?
            self.flagModifiedParVal=1
        self.widgetdef.value=newvalue                                   # copy newvalue 
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()           # write new parameter value to simulation (only for updateMode 'async')
        if not self.callback_on_radiobuttonupdate is None: self.callback_on_radiobuttonupdate()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the OptionMenu (e.g., if selecting a new value)
        if obj==None:
            obj="radiobutton"
        if obj=="radiobutton":
            self.callback_on_radiobuttonupdate=function2call
        else:
            print("Warning in IVisitRadiobuttonWidget.bind(): Unknown widget object ", obj, " in IVisitRadiobuttonWidget", self.widgetdef.name)
        return None
            
# *********************************************************************************
# class IVisitButtonWidget:
# IVisit Button to control a boolean-type simulation parameter to initiate actions
# *********************************************************************************

class IVisitButtonWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitButtonWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert isinstance(self.widgetdef,ParameterWidgetDef), "IVisitButtonWidget needs to be initialized with a ParameterWidgetDef, but self.widgetdef="+str(self.widgetdef.name)+" is not!" 
        assert self.widgetdef.type=='button', "IVisitButtonWidget needs to be initialized with a ParameterWidgetDef having type='button', but actual type="+str(self.widgetdef.type)+" is not valid!" 
        assert self.widgetdef.type_parameter in ['string','str','text'], "IVisitButtonWidget needs to be initialized with parameters of type 'string', but actual parameter_type="+str(self.widgetdef.type_parameter)+" is not valid!"
        assert isinstance(self.widgetdef.range,(tuple,list)) and len(self.widgetdef.range)==2,\
            "IVisitButtonWidget.widgetdef.range="+str(self.widgetdef.range)+" must be a list [LabelText,ButtonText] of two strings!"
        assert isinstance(self.widgetdef.value,str), "IVisitButtonWidget.widgetdef.value="+str(self.widgetdef.value)+" must have type string!"
        self.lock = threading.Lock()
        #self.label.configure(text=str(self.widgetdef.range[0]))
        self.labeltext.set(self.widgetdef.range[0])
        self.label.pack(side=LEFT)
        self.button = Button(self.labelFrame,text=str(self.widgetdef.range[1]),command=self.onPressedButton)
        self.button.pack(side=RIGHT)
        self.callback_on_buttonupdate=None
        self.setState()
        self.setValue(0)

    def setState(self):        # set object fields that are derived from widget and data definitions
        self.sizeX = self.label.winfo_reqwidth()+self.button.winfo_reqwidth()
        self.sizeY = max([self.label.winfo_reqheight(),self.button.winfo_reqheight()]) 

    def setValue(self,val):
        val=str(val)
        self.lock.acquire()                       # mutual exclusion for accessing value
        if val!=self.widgetdef.value:
            self.parent.isModified=True  # newvalue differs from previous value?
            self.flagModifiedParVal=1
        self.widgetdef.value=val                  # set new value 
        if self.widgetdef.wdefcoll.sim.updateMode!='sync': self.widgetdef.setSimParValue()           # write new parameter value to simulation (only for updateMode 'async')
        self.lock.release()

    def onPressedButton(self,e=None):
        #print("onMove...;flag=",flagIgnoreModFlag)
        #assert flagIgnoreModFlag==1
        #if(flagIgnoreModFlag<=0): 
        #    self.parent.isModified=True
        #    self.widgetdef.isModified=True
        self.setValue(1)                          # set value to 1
        if not self.callback_on_buttonupdate is None: self.callback_on_buttonupdate()

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
    def bind(self,eventType,function2call,obj=None):  # use this method to bind a function call to the button (e.g., if selecting a new value)
        if obj==None:
            obj="button"
        if obj=="button":
            self.callback_on_buttonupdate=function2call
        else:
            print("Warning in IVisitButtonWidget.bind(): Unknown widget object ", obj, " in IVisitButtonWidget", self.widgetdef.name)
        return None
            
# *********************************************************
# class IVisitTextCommentWidget:
# IVisit Comment Widget to display a comment as a textfield  
# *********************************************************

class IVisitTextCommentWidget(Frame,IVisitWidget):
    def __init__(self,parent,widgetdef_arg,className="IVisitTextCommentWidget"):
        Frame.__init__(self,master=parent)
        IVisitWidget.__init__(self,parent,widgetdef_arg,className)
        assert self.widgetdef.type=='textfield', "IVisitTextCommentWidget needs to be initialized with a CommentWidgetDef having type='textfield', but actually type="+str(self.widgetdef.type)+" is not valid!"
        if (self.widgetdef.flagDisplayName<1):
            self.labeltext.set(self.widgetdef.comment)
        else:
            n=self.widgetdef.name
            self.labeltext.set(n+'\n'+len(n)*'-'+self.widgetdef.comment)
        self.label.configure(fg=self.widgetdef.fontcolor,font=(self.widgetdef.fontname,self.widgetdef.fontsize,self.widgetdef.fontstyle))
        self.label.pack(side=TOP)
        self.setState()
        if(className=="IVisitTextCommentWidget"): self.updateData()      # derived classes may need further preprocessing before call

    def setState(self):        # set object fields that are derived from widget and data definitions
        self.sizeX, self.sizeY = self.label.winfo_reqwidth(), self.label.winfo_reqheight()
        #self.shape=np.array(self.widgetdef.data).shape
        #assert len(self.shape)>=2, "IVisitImageWidget '"+str(self.widgetdef.name)+"' is of type 'image', but is associatied to a dataarray '" + str(self.widgetdef.name_dataarray) + "' with less than 2 dimensions!" 
        #self.RGBflag=0         # default a 2D gray scale array
        #if len(self.shape)>=3: self.RGBflag=1    # 3D RGB array?
        #self.NY, self.NX = self.shape[0],self.shape[1]
        #self.CX, self.CY = int(self.widgetdef.scale*self.NX), int(self.widgetdef.scale*self.NY)
        #self.SZ          = self.CX * self.CY

    def updateData(self):
        pass

    def getWidgetSize(self):   # return (width,height) tuple
        return (self.sizeX,self.sizeY)
       
        
# *********************************************************
# class IVisitWidgetFactory:
# Factory class to create different types of IVisitWidgest
# *********************************************************
class IVisitWidgetFactory:
    def create_IVisitWidget(self,parent,widgetdef):
        assert isinstance(widgetdef,(ParameterWidgetDef,DataWidgetDef,CommentWidgetDef)), "widgetdef must be either a ParameterWidgetDef or DataWidgetDef or CommentWidgetDef!" 
        if isinstance(widgetdef,ParameterWidgetDef):
            assert widgetdef.type in ['slider','dictslider','listselection','checkbox','radiobutton','button','textfield'], \
                "ParameterWidgetDef with name="+str(widgetdef.name)+" has type="+str(widgetdef.type)+" which is not yet implemented!"
            if  (widgetdef.type=='slider'         ):
                #print("widgetdef of slider:")
                #widgetdef.printState()
                return IVisitSliderWidget(parent,widgetdef)
            elif(widgetdef.type=='dictslider'     ):
                #print("widgetdef of dictslider:")
                #widgetdef.printState()
                #exit(0)
                return IVisitDictSliderWidget(parent,widgetdef)
            elif(widgetdef.type=='listselection'  ): return IVisitListSelectionWidget(parent,widgetdef)
            elif(widgetdef.type=='checkbox'       ): return IVisitCheckboxWidget(parent,widgetdef)
            elif(widgetdef.type=='radiobutton'    ): return IVisitRadiobuttonWidget(parent,widgetdef)
            elif(widgetdef.type=='button'         ): return IVisitButtonWidget(parent,widgetdef)
            elif(widgetdef.type=='textfield'      ): return IVisitTextInputWidget(parent,widgetdef)
        if isinstance(widgetdef,DataWidgetDef):
            assert widgetdef.type in ['image','textfield'], "DataWidgetDef with name="+str(widgetdef.name)+" has type="+str(widgetdef.type)+" which is not yet implemented!"
            if   (widgetdef.type=='image'    ): return IVisitImageWidget(parent,widgetdef)
            elif (widgetdef.type=='textfield'): return IVisitTextfieldWidget(parent,widgetdef)
        if isinstance(widgetdef,CommentWidgetDef):
            assert widgetdef.type in ['textfield'], "CommentWidgetDef with name="+str(widgetdef.name)+" has type="+str(widgetdef.type)+" which is not yet implemented!"
            if(widgetdef.type=='textfield'): return IVisitTextCommentWidget(parent,widgetdef)
        return None

# ******************************************
# class IVisitRawDisplayFrame:
# Database of Widgets 
# ******************************************
class IVisitRawDisplayFrame(Frame):
    def __init__(self,parent,wdefcoll=None):
        Frame.__init__(self,master=parent,width=200,height=500)
        self.isModified=False
        self.widgets=[]
        self.dict_widgets={}
        self.data_widgets=[]
        self.par_widgets=[]
        self.com_widgets=[]
        self.wdefcoll=None
        self.createWidgets(wdefcoll)

    def createWidgets(self,wdefcoll):
        if(wdefcoll): 
            self.wdefcoll=wdefcoll
            # first remove old widgets...
            self.deleteWidgets()
            # create new widgets...
            widget_factory = IVisitWidgetFactory()
            for wdef in wdefcoll.wdefs_data+wdefcoll.wdefs_parameters+wdefcoll.wdefs_comments:
                #w=widget_factory.create_PyFeWidget(self,database,wdef)
                w=widget_factory.create_IVisitWidget(self,wdef)
                assert w, "Unable to create widget "+ wdef.name + " of type " + wdef.type 
                #sz=w.getWidgetSize()
                #self.geometry("%dx%d+%d+%d" % (sz[0],sz[1],w.posX,w.posY))
                w.place(x=wdef.pos[0],y=wdef.pos[1])
                #w.place(x=0, y=0)
                #w.pack()       
                self.widgets.append(w)
                self.dict_widgets[w.widgetdef.name]=w
                if isinstance(wdef,ParameterWidgetDef): self.par_widgets.append(w)
                if isinstance(wdef,DataWidgetDef): self.data_widgets.append(w)
                if isinstance(wdef,CommentWidgetDef): self.com_widgets.append(w)
            self.updateData()

            #database.libSim.PyFeInter_checkUpdatedParameters(0); # update parameters according to loaded slider values

    def deleteWidgets(self):
        for w in self.widgets:
            w.place_forget()
        self.widgets=[]
        self.data_widgets=[]
        self.par_widgets=[]

    def resetButtonsAndFlags(self,flagModifiedParVal=0):  # reset Button variables after each call to step() (where the values may be checked!!)
        for w in self.widgets:
            w.flagModifiedParVal=flagModifiedParVal                  # reset modified flag for ParameterWidgets
            if isinstance(w,IVisitButtonWidget): w.setValue(0)       # reset button
        
    def updateData(self):
        for w in self.data_widgets: 
            w.updateData()
        self.adaptSize()
            #w.update_idletasks()             

    def adaptSize(self):
        maxX,maxY = 100,100
        for w in self.widgets:
            xx,yy = w.widgetdef.pos[0]+w.sizeX, w.widgetdef.pos[1]+w.sizeY
            if(xx>maxX): maxX=xx
            if(yy>maxY): maxY=yy
        self.configure(width=maxX+50,height=maxY+50)
        try:
            self.master.adaptSize()
            #self.master.master.adaptSize()
            #self.master.master.master.adaptSize()
            #self.master.master.adaptSize()     # ok, this is not so elegent, making some special assumptions, but it works....
        except:
            #print "exception: no master.master.adaptSize()!!!!"
            pass
        else:
            #print "no exception: there is master.master.adaptSize()!!!!"
            pass
    
    def saveGUIDataToDatabase(self, flagForce=0):
        if (self.isModified or (flagForce>0)) and len(self.widgets)>0: self.wdefcoll.saveToDatabase()
        self.isModified=False

    def bind2Widget(self,widgetname,eventType,function2call,widgetobj=None):  # bind function2call to object obj of widget widgetname if eventType happens
        found=False
        for w in self.widgets:
            if w.widgetdef.name==widgetname:
                found=True
                break
        if found:
            res_wobj=w.bind(eventType,function2call,widgetobj)
        else:
            print("Warning in IVisitRawDisplayFrame.bind2Widget: Unknown widget ",widgetname)
            res_wobj=None
        return res_wobj

    def getWidget(self,widgetname):
        try:
            w=self.dict_widgets[widgetname]
        except KeyError:
            w=None
        finally:
            return w
        #found=False
        #for w in self.widgets:
        #    if w.widgetdef.name==widgetname:
        #        found=True
        #        break
        #if found: return w
        #else: return None

    def getFlagModified(self,widgetname=None):
        if widgetname is None: return self.isModified
        else:
            w=self.getWidget(widgetname)
            if w is None: return 0
            else: return w.flagModifiedParVal
        
    def setWidgetValue(self,widgetname,val):
        w=self.getWidget(widgetname)
        if not w is None:
            w.setValue(val)
        else:
            print("Warning in IVisitRawDisplayFrame.setWidgetValue: Unknown widget ",widgetname)
            
            
if __name__ == '__main__':
    print("\nModule test of ivisit.widgetcollection.py")
    print("-------------------------------------------\n") 
    # (i) read widgetsdefs from database
    db = sqldatabase(db_ivisit_cfg)
    sim=IVisit_Simulation()
    wdefcoll = IVisitWidgetDefCollection(db,sim)
    wdefcoll.initFromDatabase(0)
    #wdefcoll.printState()
    # (ii) create widgets and display
    root=Tk()
    f=IVisitRawDisplayFrame(root,wdefcoll)
    f.adaptSize()
    f.pack()
    wdefcoll.sim.main_init()
    wdefcoll.sim.init()
    for i in range(200): 
        wdefcoll.sim.step()
    f.updateData()
    f.update()
    root.mainloop()
 



#from Tkinter import *
from ivisit.defdb import *
from ivisit.simulation import *

# ***************************************************************************
# class ParameterWidgetDef:
# GUI-independent definition of a ParameterWidget 
#    - provides all information necessary 
#      to a create a real widget that can be displayed on the working frame 
#    - A ParameterWidgetDef thereby links a Database ParameterWidget entry 
#      to the corresponding simulation parameters of the simulation program
# ***************************************************************************
class ParameterWidgetDef:     
    def __init__(self,wdefcoll_arg,pardata=None):
        self.wdefcoll          = wdefcoll_arg   # reference to WidgetDefCollection object
        self.key               = None           # key in database table ParameterWidget or DataWidget (depending on type)
        self.name              = None           # name of the widget
        self.comment           = None           # comment 
        self.type              = None           # either 'slider' or 'listsel' or 'textfield' 
        self.range             = None           # list of [min,max,nTicks,scale] 
        self.items             = None           # list of [item1,item2,...]
        self.size              = None           # [sz_x,sz_y) size of widget
        self.pos               = None           # (x,y) position of widget
        self.value             = None           # value of the parameter
        self.key_simulation    = None           # key of the simulation
        self.key_parameter     = None           # key of the parameter
        self.name_parameter    = None           # name of the parameter
        self.comment_parameter = None           # comment of the parameter
        self.type_parameter    = None           # either 'int' or 'float' or 'text'
        self.range_parameter   = None           # either [min,max] or [val1,val2,...]
        self.listidx_parameter = None           # index of parameter value to be controlled (if parameter is list)
        self.flagListParameter = None           # if set then parameter is a list
        if pardata!=None:
            self.setData(pardata)

    def setData(self,pardata):   # set attributes according to pardata (which is a join of the columns of the tables (simulation,parameterwidget,parameter))
        # (i) do raw input from tables
        icol=self.wdefcoll.icol_s_pw_p          # set reference to index dict to pardata
        #print("icol=",icol)
        #print("pardata=",pardata)
        self.key              =pardata[icol['pw.key']]          # key of the parameter widget
        self.name             =pardata[icol['pw.name']]         # name of the parameter widget
        self.comment          =pardata[icol['pw.comment']]      # comment 
        self.type             =pardata[icol['pw.type']]         # either 'slider' or 'listsel' or 'textfield' 
        self.range            =pardata[icol['pw.range']]        # list of [min,max,nTicks,scale]
        self.items            =pardata[icol['pw.items']]        # list of [item1,item2,...]
        self.size             =pardata[icol['pw.size']]         # [sz_x,sz_y) size of widget
        self.pos              =pardata[icol['pw.pos']]          # (x,y) position of widget
        self.value            =pardata[icol['pw.value']]        # value of the parameter
        self.key_simulation   =pardata[icol['s.key']]           # key of the simulation
        self.key_parameter    =pardata[icol['p.key']]           # key of the parameter
        self.name_parameter   =pardata[icol['p.name']]          # name of the parameter
        self.comment_parameter=pardata[icol['p.comment']]       # comment of the parameter
        self.type_parameter   =pardata[icol['p.type']]          # either 'int' or 'float' or 'text'
        self.range_parameter  =pardata[icol['p.range']]         # either [min,max] or [val1,val2,...]
        self.listidx_parameter=pardata[icol['p.listidx']]       # index of parameter value to be controlled (if >=0 and parameter is list)
        self.flagListParameter=0                                # default no list parameter
        # (ii) refined post-processing
        self.type=self.type.strip()                             # strip white space
        self.type_parameter=self.type_parameter.strip()         # strip white space
        self.range=self.wdefcoll.db.parseStringAsList(self.range,self.type_parameter,[0,1,1,1])                   # finalize self.range as list of ints or floats or strings
        self.items=self.wdefcoll.db.parseStringAsList(self.items,'string',[0,1])                                  # finalize self.items as list of strings
        self.size=self.wdefcoll.db.parseStringAsList(self.size,'int',[0,0])                                       # finalize self.size as list of ints 
        self.range_parameter=self.wdefcoll.db.parseStringAsList(self.range_parameter,self.type_parameter,[0,1])   # finalize self.range_parameter as list of ints or floats or strings
        self.pos=self.wdefcoll.db.parseStringAsList(self.pos,'int',[0,0])                                         # finalize self.pos as list of ints [x,y]
        if(self.type_parameter=='int'):
            try:
                self.value=int(self.value)
            except ValueError as e:
                print("Warning: cannot convert self.value=",self.value," to int for parameter widget with name=",self.name)
                self.value=0
                pass
        else:
            if(self.type_parameter=='float'):
                self.value=float(self.value)
        if self.type=='dictslider':
            # special handling for dictsliders
            self.items=[str(s).strip() for s in self.items]   # self.items has format: [nitems,idxselecteditem,  itemname1,type1,  itemname2,typ2, ...] (length 2+nitems*2)
            self.items[0]=asNumber(self.items[0],'int',0)
            self.items[1]=asNumber(self.items[1],'int',0)
            nitems=self.items[0]
            self.itemtypes=[self.items[2+i*2+1] for i in range(nitems)]         # list of item types (either 'int' or 'float' for each item)
            self.range          =[str(s).strip() for s in self.range]           # self.range has format: [nitems, itemname1,min1,max1,nticks1,scale1, itemname2,min2,max2,nticks2,scale2, ...] (length 1+nitems*5 
            self.range_parameter=[str(s).strip() for s in self.range_parameter] # self.range_parameter has format: [nitems, key1,type1,min1,max1, key2,type2,min2,max2, ...] (length 1+nitems*4
            self.range[0]=asNumber(self.range[0],'int',0)
            self.range_parameter[0]=asNumber(self.range_parameter[0],'int',0)
            for i in range(nitems):
                ibase=1+i*5
                ibasep=1+i*4
                self.range[ibase+3]=asNumber(self.range[ibase+3],'int',0)   # handle nticksi
                tp=self.itemtypes[i]                                        # numerical type (either 'int' or 'float')
                self.range[ibase+1],self.range[ibase+2],self.range[ibase+4]  =asNumber(self.range[ibase+1],tp,0),asNumber(self.range[ibase+2],tp,1),asNumber(self.range[ibase+4],tp,1)  # handle min1,max1,scale1
                self.range_parameter[ibasep+2],self.range_parameter[ibasep+3]=asNumber(self.range_parameter[ibasep+2],tp,0),asNumber(self.range_parameter[ibasep+3],tp,1)               # handle min1,max1
        # finalize
        p=self.wdefcoll.getSimParameterValue(self.name_parameter,0)
        if isinstance(p, (list, tuple)): self.flagListParameter=1

    def getSimParValue(self,default_value=None):      # return current parameter value from simulation program
        p=self.wdefcoll.getSimParameterValue(self.name_parameter,default_value)
        if self.flagListParameter>0:
            assert isinstance(self.value, (list, tuple)),"Simulation parameter name_parameter="+self.name_parameter+" is neither list nor tuple, but ParameterWidgetDef "+\
                   self.name+" says flagListParameter="+str(self.flagListParameter)
            assert (self.listidx_parameter>=0)and(self.listidx_parameter<len(p),"listidx_parameter="+str(self.listidx_parameter)+" is out of allowed range >=0 and <len(p)="+str(len(p)))
            p=p[self.listidx_parameter]
        else:
            assert not isinstance(self.value, (list, tuple)),"Simulation parameter name_parameter="+self.name_parameter+" is list or tuple, but ParameterWidgetDef "+\
                   self.name+" says flagListParameter="+str(self.flagListParameter)
        return p

    def setSimParValue(self):                         # write parameter value to simulation program 
        if(self.flagListParameter):
            p=self.wdefcoll.getSimParameterValue(self.name_parameter)   # get reference to simulation parameter list
            assert isinstance(p, (list, tuple)),"Simulation parameter name_parameter="+self.name_parameter+" is neither list nor tuple, but ParameterWidgetDef "+\
                   self.name+" says flagListParameter="+str(self.flagListParameter)
            assert (self.listidx_parameter>=0)and(self.listidx_parameter<len(p)),"listidx_parameter="+str(self.listidx_parameter)+" is out of allowed range >=0 and <len(p)="+str(len(p))
            p[self.listidx_parameter]=self.value
        elif self.type=='dictslider':
            assert hasattr(self, 'ref_to_dictsliderwidget'), "Ooops! Expected that ParameterWidgetDef object of type dicslider should have attribute ref_to_dictsliderwidget!!"\
                +"\nIf this happens try to implement access to dict value (extracting from string list) as in the constructor of IVisitDictSliderWidget!!"
            p=self.wdefcoll.getSimParameterValue(self.name_parameter)
            assert isinstance(p,dict),"Expected parameter p with name_parameter="+str(self.name_parameter)+" to be a dict for dictslider "+str(self.name)+", but p="+str(p)
            ds=self.ref_to_dictsliderwidget   # DIRTY!!!! See widgets.py, class IVisitDictSliderWidget
            for i in range(ds.nitems): p[ds.itemkeys[i]]=ds.values[i]
        else:
            self.wdefcoll.setSimParameterValue(self.name_parameter,self.value) 
        
    def saveWidgetDataToDatabase(self):              # save the parameter value and position from widgetdef (that may have been manipulated in GUI) to database
        self.wdefcoll.db.simple_update_byPKEY('parameterwidget',['value','pos'],[str(self.value),str(self.pos)],['key'],[str(self.key)])

    def printState(self,indent=""):
        print(indent+"ParameterWidgetDef "+str(self.name)+" :")
        print(indent+"   key               ="+str(self.key))
        print(indent+"   comment           ="+str(self.comment))
        print(indent+"   type              ="+str(self.type))
        print(indent+"   range             ="+str(self.range))
        print(indent+"   items             ="+str(self.items))
        print(indent+"   size              ="+str(self.size))
        print(indent+"   pos               ="+str(self.pos))
        print(indent+"   value             ="+str(self.value))
        print(indent+"   key_simulation    ="+str(self.key_simulation))
        print(indent+"   key_parameter     ="+str(self.key_parameter))
        print(indent+"   name_parameter    ="+str(self.name_parameter))
        print(indent+"   comment_parameter ="+str(self.comment_parameter))
        print(indent+"   type_parameter    ="+str(self.type_parameter))
        print(indent+"   range_parameter   ="+str(self.range_parameter))
        print(indent+"   listidx_parameter ="+str(self.listidx_parameter))
        print(indent+"   flagListParameter ="+str(self.flagListParameter))
        print(indent+"   data_value        ="+str(self.type))

# ***************************************************************************
# class DataWidgetDef:
# GUI-independent definition of a DataWidget 
#    - provides all information necessary 
#      to a create a real widget that can be displayed on the working frame 
#    - A DataWidgetDef thereby links a Database DataWidget entry 
#      to the corresponding simulation DataArrays of the simulation program
# ***************************************************************************
class DataWidgetDef:     
    def __init__(self,wdefcoll_arg,datadata=None):
        self.wdefcoll          = wdefcoll_arg   # reference to WidgetDefCollection object
        self.key               = None           # key of the data widget
        self.name              = None           # name of the data widget
        self.comment           = None           # comment 
        self.type              = None           # either 'image' or 'textfield' 
        self.range             = None           # list of [min,max] 
        self.pos               = None           # (x,y) position of widget
        self.scale             = None           # scale of the data widget 
        self.data              = None           # reference to the data object (typically a numpy array)
        self.key_simulation    = None           # key of the simulation
        self.key_dataarray     = None           # key of the dataarray
        self.name_dataarray    = None           # name of the dataarray
        self.comment_dataarray = None           # comment of the dataarray
        self.type_dataarray    = None           # either 'int' or 'float' or 'binary' or 'text'
        self.range_dataarray   = None           # [min,max] 
        self.flagRefData       = None           # if flag>0 then self.data is a reference (to an array); otherwise it is a non-referential data (e.g.,an integer)
        if datadata!=None:
            self.setData(datadata)

    def setData(self,datadata):   # set attributes according to datadata (which is a join of the columns of the tables (simulation,datawidget,dataarray))
        # (i) do raw input from tables
        icol=self.wdefcoll.icol_s_dw_da          # set reference to index dict to datadata
        self.key              =datadata[icol['dw.key']]        # key of the data widget
        self.name             =datadata[icol['dw.name']]       # name of the data widget
        self.comment          =datadata[icol['dw.comment']]    # comment 
        self.type             =datadata[icol['dw.type']]       # either 'image' or 'textfield' 
        self.range            =datadata[icol['dw.range']]      # list of [min,max] 
        self.pos              =datadata[icol['dw.pos']]        # (x,y) position of widget
        self.scale            =datadata[icol['dw.scale']]      # scale of the data widget 
        self.data             =None                            # reference to the data object (typically a numpy array)
        self.key_simulation   =datadata[icol['s.key']]         # key of the simulation
        self.key_dataarray    =datadata[icol['da.key']]        # key of the dataarray
        self.name_dataarray   =datadata[icol['da.name']]       # name of the dataarray
        self.comment_dataarray=datadata[icol['da.comment']]    # comment of the dataarray
        self.type_dataarray   =datadata[icol['da.type']]       # either 'int' or 'float' or 'binary' or 'text'
        self.range_dataarray  =datadata[icol['da.range']]      # list of [min,max] 
        self.flagRefData      =0                               # default no ref parameter 
        # (ii) refined post-processing
        self.type=self.type.strip()                            # strip white space
        self.type_dataarray=self.type_dataarray.strip()        # strip white space
        self.range=self.wdefcoll.db.parseStringAsList(self.range,self.type_dataarray,[0,1])                      # finalize self.range as list of ints or floats 
        self.range_dataarray=self.wdefcoll.db.parseStringAsList(self.range_dataarray,self.type_dataarray,[0,1])  # finalize self.range_parameter as list of ints or floats or strings
        self.pos=self.wdefcoll.db.parseStringAsList(self.pos,'int',[0,0])                                        # finalize self.pos as list of ints [x,y]
        self.scale=float(self.scale)                           # convert to float
        self.setDataFromSimulation(0)                          # set self.data by simulation data
        self.flagRefData=0                                     # test if self.data has reference-type (1) or not (0)
        if isinstance(self.data, (list, tuple, np.ndarray)):   # ditto...
            self.flagRefData=1

    def setDataFromSimulation(self,default_value=None):        # get current simulation data from simulation program and set self.data accordingly
        self.data=self.wdefcoll.getSimData(self.name_dataarray,default_value)

    def saveWidgetDataToDatabase(self):              # save the widget position from widgetdef (that may have been manipulated in GUI) to database
        self.wdefcoll.db.simple_update_byPKEY('datawidget',['pos'],[str(self.pos)],['key'],[str(self.key)])

    def printState(self,indent=""):
        print(indent+"DataWidgetDef "+str(self.name)+" :")
        print(indent+"   key               ="+str(self.key))
        print(indent+"   comment           ="+str(self.comment))
        print(indent+"   type              ="+str(self.type))
        print(indent+"   range             ="+str(self.range))
        print(indent+"   pos               ="+str(self.pos))
        print(indent+"   scale             ="+str(self.scale))
        print(indent+"   data              ="+str(self.data))
        print(indent+"   key_simulation    ="+str(self.key_simulation))
        print(indent+"   key_dataarray     ="+str(self.key_dataarray))
        print(indent+"   name_dataarray    ="+str(self.name_dataarray))
        print(indent+"   comment_dataarray ="+str(self.comment_dataarray))
        print(indent+"   type_dataarray    ="+str(self.type_dataarray))
        print(indent+"   range_dataarray   ="+str(self.range_dataarray))
        print(indent+"   flagRefData       ="+str(self.flagRefData))


# ***************************************************************************
# class CommentWidgetDef:
# GUI-independent definition of a CommentWidget 
#    - provides all information necessary 
#      to a create a real widget that can be displayed on the working frame 
#    - Note that a CommentWidgetDef is not linked to a database entry! 
# ***************************************************************************
class CommentWidgetDef:     
    def __init__(self,wdefcoll_arg,comdata=None):
        self.wdefcoll          = wdefcoll_arg   # reference to WidgetDefCollection object
        self.key               = None           # key of the comment widget
        self.name              = None           # name of the comment widget
        self.comment           = None           # comment 
        self.type              = None           # either 'textfield' or ??? 
        self.fontname          = None           # fontname (e.g., Arial)  
        self.fontsize          = None           # fontsize (e.g., 10)
        self.fontstyle         = None           # fontstyle (e.g., Normal)  
        self.fontcolor         = None           # fontcolor (e.g., black) 
        self.flagDisplayName   = None           # if flag=1 then display name on frame; if flag=0 then do not display
        self.pos               = None           # (x,y) position of widget
        self.key_simulation    = None           # key of the simulation
        if comdata!=None:
            self.setData(comdata)

    def setData(self,comdata):   # set attributes according to comdata (which is a join of the columns of the tables (simulation,datawidget,dataarray))
        # (i) do raw input from tables
        icol=self.wdefcoll.icol_s_cw          # set reference to index dict to comdata
        self.key              =comdata[icol['cw.key']]             # key of the comment widget
        self.name             =comdata[icol['cw.name']]            # name of the comment widget
        self.comment          =comdata[icol['cw.comment']]         # comment 
        self.type             =comdata[icol['cw.type']]            # either 'image' or 'textfield' 
        self.fontname         =comdata[icol['cw.fontname']]        # fontname (e.g., Arial)  
        self.fontsize         =comdata[icol['cw.fontsize']]        # fontsize (e.g., 10)
        self.fontstyle        =comdata[icol['cw.fontstyle']]       # fontstyle (e.g., Normal)  
        self.fontcolor        =comdata[icol['cw.fontcolor']]       # fontcolor (e.g., black) 
        self.flagDisplayName  =comdata[icol['cw.flagDisplayName']] # if flag=1 then display name on frame; if flag=0 then do not display
        self.pos              =comdata[icol['cw.pos']]             # (x,y) position of widget
        self.key_simulation   =comdata[icol['s.key']]              # key of the simulation
        # (ii) refined post-processing
        self.type=self.type.strip()                                # strip white space
        self.fontsize=int(self.fontsize)                           # convert to int
        self.flagDisplayName=int(self.flagDisplayName)             # convert to int
        self.pos=self.wdefcoll.db.parseStringAsList(self.pos,'int',[0,0]) # finalize self.pos as list of ints [x,y]

    def saveWidgetDataToDatabase(self):              # save the widget position from widgetdef (that may have been manipulated in GUI) to database
        self.wdefcoll.db.simple_update_byPKEY('commentwidget',['pos'],[str(self.pos)],['key'],[str(self.key)])

    def printState(self,indent=""):
        print(indent+"CommentWidgetDef "+str(self.name)+" :")
        print(indent+"   key               ="+str(self.key))
        print(indent+"   comment           ="+str(self.comment))
        print(indent+"   type              ="+str(self.type))
        print(indent+"   fontname          ="+str(self.fontname))
        print(indent+"   fontsize          ="+str(self.fontsize))
        print(indent+"   fontstyle         ="+str(self.fontstyle))
        print(indent+"   fontcolor         ="+str(self.fontcolor))
        print(indent+"   flagDisplayName   ="+str(self.flagDisplayName))
        print(indent+"   pos               ="+str(self.pos))
        print(indent+"   key_simulation    ="+str(self.key_simulation))


# *********************************************************************
# *********************************************************************
# class IVisitWidgetDefCollection:
# Collection of widget definitions, e.g., read from a ivisit database
#
# The purpose of this class is to link database widget objects to 
# their corresponding simulation entities (like parameters and data) 
# *********************************************************************
# *********************************************************************
class IVisitWidgetDefCollection:
    def __init__(self, db_arg, sim_arg):
        # connect to database and simulation program
        self.db  = db_arg     # Reference to IVisit database
        self.sim = sim_arg    # Reference to simulation object (having a field 'parameters')
        self.simpars=self.sim.parameters_dict  # reference to working copy of the simulation parameters
        self.simpars_original=self.sim.parameters_dict_original # reference to (deep) copy of the original simulation parameters (as defined in class struct IVisit_Parameters)
        self.dataarrays = self.sim.data_dict # reference to data fields of the simulation program
        # simulation key and name (refers to database table 'simulation')
        self.simulation_key = None  # simulation.key
        self.simulation_name = ""   # simulation.name
        # simulation control parameters        
        self.simulation_simsteps_per_frame = 1    # number of simulation steps per displayed frame (i.e., #Felix steps per call to interface step() 
        self.simulation_frames_per_step    = 1    # number of display frames per call to onStep()  
        self.simulation_delay_per_step     = 1    # delay (in msec) per call to self.onStep() (e.g., after run or cont)
        # dicts for easy access of table columns
        self.icol_s=icol_simulation
        l1,l2,l3=['s.'+c for c in tb_simulation_cfg.col_names], ['pw.'+c for c in tb_parameterwidget_cfg.col_names], ['p.'+c for c in tb_parameter_cfg.col_names]
        l=l1+l2+l3
        self.icol_s_pw_p={l[i]:i for i in range(len(l))}    # e.g., icol_s_pw_p['pw.key'] is index of parameterwidget.key within join of simulation,parameterwidget,parameter
        l2,l3=['dw.'+c for c in tb_datawidget_cfg.col_names], ['da.'+c for c in tb_dataarray_cfg.col_names]
        l=l1+l2+l3
        self.icol_s_dw_da={l[i]:i for i in range(len(l))}   # e.g., icol_s_dw_da['dw.key'] is index of datawidget.key within join of simulation,datawidget,dataarray
        l4=['cw.'+c for c in tb_commentwidget_cfg.col_names]
        l=l1+l4
        self.icol_s_cw={l[i]:i for i in range(len(l))}      # e.g., icol_s_dw_da['dw.key'] is index of datawidget.key within join of simulation,datawidget,dataarray
        # very database (list of IVisitWidgetDef objects)
        self.wdefs_parameters = []     # list of all ParameterWidget definitions
        self.wdefs_data       = []     # list of all DataWidget definitions
        self.wdefs_comments   = []     # list of all CommentWidget definitions

    def initFromDatabase(self,sim_key):    # create and initialize all widgetdefs from the database for simulation with key sim_key
        # (i) read in data of simulation table
        res = self.db.simple_select('*','simulation',where='key='+str(sim_key))[0]
        self.simulation_key               =res[icol_simulation['key']]
        self.simulation_name              =res[icol_simulation['name']]
        self.simulation_comment           =res[icol_simulation['comment']]
        self.simulation_date_init         =res[icol_simulation['date_init']]
        self.simulation_date_lastmod      =res[icol_simulation['date_lastmod']]
        self.simulation_simsteps_per_frame=res[icol_simulation['simsteps_per_frame']]
        self.simulation_frames_per_step   =res[icol_simulation['frames_per_step']]
        self.simulation_delay_per_step    =res[icol_simulation['delay_per_step']]
        # (ii) read in parameter widgets
        joinon_clause = ['parameterwidget.key_simulation=simulation.key','parameterwidget.key_parameter=parameter.key']
        res = self.db.simple_select('*',['simulation','parameterwidget','parameter'],joinon=joinon_clause,where='simulation.key='+str(sim_key))
        self.wdefs_parameters = [ParameterWidgetDef(self,d)   for d in res]
        # (iii) read in data widgets
        joinon_clause = ['datawidget.key_simulation=simulation.key','datawidget.key_dataarray=dataarray.key']
        res = self.db.simple_select('*',['simulation','datawidget','dataarray'],joinon=joinon_clause,where='simulation.key='+str(sim_key))
        self.wdefs_data = [DataWidgetDef(self,d)   for d in res]
        # (iv) read in comment widgets
        joinon_clause = ['commentwidget.key_simulation=simulation.key']
        res = self.db.simple_select('*',['simulation','commentwidget'],joinon=joinon_clause,where='simulation.key='+str(sim_key))
        self.wdefs_comments = [CommentWidgetDef(self,d)   for d in res]

    def saveToDatabase(self):  # save all data from widgetdefs (that may have been manipulated by GUI) to database 
        for wdef in self.wdefs_parameters: wdef.saveWidgetDataToDatabase()
        for wdef in self.wdefs_data      : wdef.saveWidgetDataToDatabase()
        for wdef in self.wdefs_comments  : wdef.saveWidgetDataToDatabase()

    def getSimParameterValue(self,parname,default_value=None):    # get simulation parameter value from simulation program
        try:
            p=getattr(self.sim.parameters,parname)
        except AttributeError as e:
            if(default_value!=None):
                print("Warning: AttributeError in IVisitWidgetDefCollection.getSimParameterValue(...): parameter with parname=",parname,\
                      " seems not to exist in simulation program (simpars=",self.simpars,")! Returning default_value=",default_value) 
                p=default_value
            else:
                raise e
        return p

    def setSimParameterValue(self,parname,value):                # set simulation parameter value in simulation program
        try:
            p=setattr(self.sim.parameters,parname,value)
        except AttributeError as e:
            print("Error: AttributeError in IVisitWidgetDefCollection.setSimParameterValue(...): parameter with parname=",parname,\
                  " seems not to exist in simulation program (simpars=",self.simpars,")! Returning default_value=",default_value) 
            raise e

    def getSimData(self,dataarray_name,default_value=None):      # get simulation data field from simulation program
        try:
            d=getattr(self.sim.data,dataarray_name)
            #print("getSimData, dataarray_name=",dataarray_name," d=",d)
        except AttributeError as e:
            if(default_value!=None):
                d=default_value
                print("Warning: AttributeError in IVisitWidgetDefCollection.getSimData(...): Data array with dataarray_name=",dataarray_name,\
                      " seems not to exist in simulation program (datarrays=",self.dataarrays,")! Returning default_value=",default_value) 
                d=default_value
            else:
                raise e
        #print("finally: getSimDAta, dataarray_name=",dataarray_name," d=",d)
        return d

    def setAllSimulationParameters(self):   # write all parameter values to simulation program 
        for pw in self.wdefs_parameters:    # loop over all parameter widgets
            pw.setSimParValue()             # write parameter value to simulation program 

    def setAllSimulationDataArrays(self):   # get all data arrays from simulation program and write to data widgets  
        for dw in self.wdefs_data:          # loop over all data widgets
            dw.setDataFromSimulation()      # get data from simuation program and write to data widget  

    def printState(self):
        print("IVisitWidgetDefCollection "+self.db.filename+":")
        print("   Simulation Program sim.name   = ",self.sim.name)
        print("   simulation_key                = ",self.simulation_key)
        print("   simulation_name               = ",self.simulation_name)
        print("   simulation_simsteps_per_frame = ",self.simulation_simsteps_per_frame)
        print("   simulation_frames_per_step    = ",self.simulation_frames_per_step)
        print("   simulation_delay_per_step     = ",self.simulation_delay_per_step)
        print("   There are "+str(len(self.wdefs_parameters))+" parameter widgets:")
        for pw in self.wdefs_parameters: pw.printState("   ") 
        print("   There are "+str(len(self.wdefs_data))+" data widgets:")
        for dw in self.wdefs_data: dw.printState("   ") 
        print("   There are "+str(len(self.wdefs_comments))+" comment widgets:")
        for cw in self.wdefs_comments: cw.printState("   ") 
        
'''
        colnames = tb_parameterwidget ['ParameterWidget.key','ParameterWidget.name','ParameterWidget.comment','Parameter
        self.db.simple_select(['Parameter'])

        fh=open(fname,'r')
        state=0        # 0=try to read simulationfilename; 1=try to read widget definition; 2=try to read widget def parameter
        entry=PyFeWidgetDef()
        parlist_name = []
        parlist_val  = []
        newdb=PyFeWidgetDefDatabase()
        while(1) :
            line = fh.readline()
            if ((not line)or(line.isspace()))and(state==2):
                state=1
                entry.parameter_names = parlist_name
                entry.parameter_values = parlist_val
                newdb.db.append(entry)
                entry=PyFeWidgetDef()
            if not line:
                break
            if line.isspace():
                continue
            s=line.split()
            s=self.removeComments(s)      # ignore comments
            if len(s)==0:            # was the line only comments?
                continue
            if state==0:
                # scan simulation_name   felixsteps_per_frame   frames_per_step   delay_per_step
                assert (len(s)<=4)and(len(s)>=1), "ParseError: Expected simulation_name  felixsteps_per_frame frames_per_step delay_per_step"
                newdb.simulation=s[0]
                if(len(s)>1): newdb.felixsteps_per_frame=int(s[1])
                if(len(s)>2): newdb.frames_per_step=int(s[2])
                if(len(s)>3): newdb.delay_per_step=int(s[3])
                state=1
                continue
            if state==1:
                # scan basic widget definition
                assert len(s)==4, "ParseError: Expected widget definitions (data_name widget_type posx posy)"
                entry.data_name=s[0]
                entry.widget_type=s[1]
                entry.pos = (int(s[2]),int(s[3]))
                state=2
                parlist_name=[]
                parlist_val=[]
                continue
            if state==2:
                # scan widget definition parameter
                assert len(s)==2, "Parse Error: Expected widget definition parameter (parameter_name parameter_value)"
                parlist_name.append(s[0])
                parlist_val.append(s[1])
        # after end-of-file, state should be 1!!!
        assert state==1, "Parse error: Unexpected end of file " + fname 
        self.simulation = newdb.simulation
        self.db         = newdb.db

    def writeToFile(self,fname):
        fh=open(fname,'w')
        fh.write(self.simulation+"\n")
        fh.write("\n")
        for entry in self.db:
            fh.write(entry.data_name + " " + entry.widget_type + " " + str(entry.pos[0]) + " " + str(entry.pos[1]) + "\n")
            assert len(entry.parameter_names)==len(entry.parameter_values), "lists of parameter names and values must have same length!"
            for i in range(len(entry.parameter_names)):
                fh.write("   "+entry.parameter_names[i]+" "+entry.parameter_values[i]+"\n")
            fh.write("\n")
            
    def printPyFeWidgetDefDatabase(self): # print database
        print "PyFeWidgetDefDatabase (size=%i): " % len(self.db)
        print "Simulation = ", self.simulation
        for entry in self.db: 
            print "data_name =", entry.data_name, "  widget_type =", entry.widget_type, "  pos_x =", entry.pos[0], "  pos_y =", entry.pos[1]
            assert len(entry.parameter_names)==len(entry.parameter_values), "lists of parameter names and values must have same length!"
            for i in range(len(entry.parameter_names)):
                print "parameter_name =", entry.parameter_names[i], "  parameter_value =", entry.parameter_values[i]

'''

if __name__ == '__main__':
    print("\nModule test of ivisit.widgetcollection.py")
    print("-------------------------------------------\n") 
    db = sqldatabase(db_ivisit_cfg)
    sim=IVisit_Simulation()
    wdefcoll = IVisitWidgetDefCollection(db,sim)
    wdefcoll.initFromDatabase(0)
    wdefcoll.printState()

    s1='[1,2.51,3,4,5]'
    print("s1=",s1," list of int = ",db.parseStringAsList(s1,'int'))
    print("s1=",s1," list of float = ",db.parseStringAsList(s1,'float'))
    s2="[1,2.51,'3',4,'5']"
    print("s2=",s2," list of string = ",db.parseStringAsList(s1,'string'))

    #db.print_database(1)

    #fname="../example/exampleSS.gui"
    #print("Connecting to default database")
    #db = PyFeWidgetDefDatabase()
    #db.readFromFile(fname)
    #db.printPyFeWidgetDefDatabase()
    #db.writeToFile(fname+".bak")    
    



import os.path
from datetime import datetime
from ivisit.defdb import *
from supy.sqldatabase import *
from supy.utilities import *

# ***************************************************************************
# class IVisitParser:
# Parsing of simulation parameters and data to be displayed  
#    - provides all information necessary 
#      to a create a real widget that can be displayed on the working frame 
#    - A ParameterWidgetDef thereby links a Database ParameterWidget entry 
#      to the corresponding simulation parameters of the simulation program
# ***************************************************************************
class IVisitParser:
    PARSEMODE_OnlyNewWidgets       ='Parse ONLY NEW WIDGETS'
    PARSEMODE_UpdateOld            ='Also UPDATE OLD WIDGETS'
    PARSEMODE_UpdateOldExceptValues='Also UPDATE OLD WIDGETS except VALUES'
    PARSEMODE_OPTIONS=[PARSEMODE_OnlyNewWidgets,PARSEMODE_UpdateOld,PARSEMODE_UpdateOldExceptValues]
    parsemode_helptext= 'Parse ONLY NEW WIDGETS: Keep all existing widgets unmodified\n'+\
                        'Also UPDATE OLD WIDGETS: Overwrite all existing widgets with parsed information\n'+\
                        'Also UPDATE OLD WIDGETS except VALUES: Overwrite all existing widgets with parsed information, but keep widget values\n'

    def __init__(self,filepath,db=None,parsemode=PARSEMODE_UpdateOldExceptValues,sim_name=None):
        self.db=db                # reference to sqldatabase
        self.parsemode=parsemode  # parsing mode
        self.sim_name=sim_name    # simulation name for which parameters should be parsed
        self.parse_lines_of_simulations(filepath)
        #print(self.lines_of_simulations)
        #print(self.linenumbers_of_simulations)
        self.nSimulations = len(self.linenumbers_of_simulations)
        self.tb_simulation_entries=self.nSimulations*[[]]
        self.tb_parameter_entries=self.nSimulations*[[]]
        self.tb_dataarray_entries=self.nSimulations*[[]]
        self.tb_parameterwidget_entries=self.nSimulations*[[]]
        self.tb_datawidget_entries=self.nSimulations*[[]]
        self.tb_commentwidget_entries=self.nSimulations*[[]]
        self.i_sim=0   # initialize simulation index (will be updated by new_simulation...)
        for i in range(self.nSimulations):
            self.parse_simulation(self.lines_of_simulations[i],self.linenumbers_of_simulations[i])

    def parse_lines_of_simulations(self,filepath):
        lines_of_simulations = []               # list of all lists of lines_of_simulations ...
        linenumbers_of_simulations = []         # ... corresponding lists of line numbers
        lines_of_simulation = []                # list of all all text lines that correspond to the current simulation and ...
        linenumbers_of_simulation = []          # ... corresponding list of line numbers (referring to the filepath file) 
        if os.path.isfile(filepath):
            with open(filepath,'r') as f:
                linenumber=0  # initialize line number
                n_simu=0      # initialize number of simulations to be parsed
                for line in f:
                    linenumber=linenumber+1          # number of the line that is currently parsed
                    l=line.strip()                   # remove white space from left and right of line
                    i_ivisit = l.find(r"@IVISIT:")   # index of beginning of @IVISIT definition (or <0)
                    if i_ivisit>=0:                  # is there a @IVISIT definition?
                        assert len(l)>(i_ivisit+8),"IVisitParser Syntax Error in File "+str(filepath)+", line "+str(linenumber)\
                            +": @IVISIT: directive must be followed by some keyword (e.g., SIMULATION, SLIDER, LISTSEL, CHECKBOX, RADIOBUTTON, TEXT_IN, IMAGE, TEXT_OUT)"
                        l=l[(i_ivisit+8):]           # if so, then take rest of line
                        i_comments = l.find(r"#")    # index of beginning of a comment (to be skipped)
                        if i_comments>=0: l=l[0:i_comments]  # skip behind the comment '#'
                        l=l.strip()                  # remove again white space from left and right of line
                        if l.find(r"SIMULATION")==0: # beginning of a new simulation definition?
                            # yes, new simulation definition
                            if len(lines_of_simulation)>0:  # add current line to list of lines and ...
                                lines_of_simulations      .append(lines_of_simulation)
                                linenumbers_of_simulations.append(number_of_lines_of_simulation)
                            lines_of_simulation       = [] # ... reset lists
                            linenumbers_of_simulation = [] # ... ditto
                            n_simu                    = n_simu+1
                        # another line belonging to the current simulation definition
                        assert n_simu>0, "IVisitParser Syntax Error in File "+str(filepath)+", line "+str(linenumber)\
                            +": First @IVISIT definition must begin with @IVISIT:SIMULATION <Name>"
                        lines_of_simulation      .append(l)
                        linenumbers_of_simulation.append(linenumber)
                if len(lines_of_simulation)>0:
                    lines_of_simulations      .append(lines_of_simulation)
                    linenumbers_of_simulations.append(linenumbers_of_simulation)
        self.lines_of_simulations       = lines_of_simulations
        self.linenumbers_of_simulations = linenumbers_of_simulations
        self.filepath=filepath

    def parse_simulation(self,lines_of_simulation, linenumbers_of_simulation):
        assert len(lines_of_simulation)==len(linenumbers_of_simulation),"IVisitParser::parse_simulation(lines_of_simulation,linenumbers_of_simulation): "\
            +"Two parameters must be lists of same length, but len(los)="+str(len(lines_of_simulation))+", len(lnos)="+str(len(linenumbers_of_simulation))+" !"
        i=0
        while i<len(lines_of_simulation):
        #for i in range(len(lines_of_simulation)):
            l,ln=lines_of_simulation[i].strip(), str(linenumbers_of_simulation[i])    # get next line
            terms=l.split('&')
            terms=[t.strip() for t in terms]
            nterms = len(terms)
            str_parseerror = "IVisitParser::parse_simulation(.): Parse error in line "+ln+" of file "+self.filepath+": "
            assert nterms>=1, str_parseerror\
                +"Expected key word (either SIMULATION, SLIDER, LISTSEL, TEXT_IN, IMAGE, or TEXT_OUT), but got nothing: terms="+str(terms)
            if terms[0]=="SIMULATION":
                assert i==0, str_parseerror+"SIMULATION DIRECTIVE must occur only once per simulation! But here SIMULATION directive occured i="+str(i)+" times!"
                assert nterms==2, str_parseerror+"Syntax Error in SIMULATION DIRECTIVE, nterms="+str(nterms)+"! Instead please use syntax @IVISIT:SIMULATION § <simulation_name>"
                self.new_simulation(terms[1])
            elif terms[0]=="SLIDER":
                assert nterms==8, str_parseerror+"Syntax Error in SLIDER DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:SLIDER & <slidername> & <size> & <range> & <parametername> & <listindex> & <type> & <value>"
                self.new_slider(terms[1],terms[2],terms[3],terms[4],terms[5],terms[6],terms[7])
            elif terms[0]=="DICTSLIDER":
                assert nterms==5, str_parseerror+"Syntax Error in DICTSLIDER DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:DICTSLIDER & <dictslidername> & <size> & <parametername> & <idx_selecteditem" 
                dictkeys=parseStringAsList(terms[3],'string','ERROR')
                assert dictkeys!='ERROR', str_parseerror+"Syntax Error in DICTSLIDER DIRECTIVE @IVISIT:DICTSLIDER & <dictslidername> & <size> & <parametername>: Could not read in dictkeys! Must be list of strings!"
                j=0
                itemnames=[]
                ranges=[]
                dictkeys=[]
                types=[]
                values=[]
                while i+j+1<len(lines_of_simulation):
                    j+=1
                    l_,ln_=lines_of_simulation[i+j].strip(), str(linenumbers_of_simulation[i+j])    # get next line
                    terms_=l_.split('&')
                    terms_=[t.strip() for t in terms_]
                    nterms_ = len(terms_)
                    str_parseerror_ = "IVisitParser::parse_simulation(.): Parse error in line "+ln_+" of file "+self.filepath+": "
                    if terms_[0]!="DICTSLIDERITEM":
                        break
                    else:
                        #print("nterm_=",nterms_)
                        assert nterms_==6, str_parseerror_+"Syntax Error in DICTSLIDERITEM DIRECTIVE, nterms="+str(nterms_)+"! "\
                            +"Instead please use syntax @IVISIT:DICTSLIDERITEM & <itemname> & <range> & <dictkey> & <type> & <value>"
                        itemnames,ranges,dictkeys,types,values = itemnames+[terms_[1]],ranges+[terms_[2]],dictkeys+[terms_[3]],types+[terms_[4]],values+[terms_[5]]
                nitems=len(itemnames)
                assert nitems>0, str_parseerror+"Syntax Error in DICTSLIDER DIRECTIVE: No DICTSLIDERITEM directives found! There should be at least one DICTSLIDERITEM following DICTSLIDER directive!"\
                    +"Use syntax @IVISIT:DICTSLIDERITEM & <itemname> & <range> & <dictkey> & <type> & <value>"
                self.new_dictslider(terms[1],terms[2],terms[3],terms[4],nitems,itemnames,ranges,dictkeys,types,values)
                i+=nitems
            elif terms[0]=="LISTSEL":
                assert nterms==8, str_parseerror+"Syntax Error in LISTSEL DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:LISTSEL & <listsel_name> & <size> & <list_of_options> & <parametername> & <listindex> & <type> & <value>"
                self.new_listsel(terms[1],terms[2],terms[3],terms[4],terms[5],terms[6],terms[7])
            elif terms[0]=="CHECKBOX":
                assert nterms==5, str_parseerror+"Syntax Error in CHECKBOX DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:CHECKBOX & <checkbox_name> & <list_of_options> & <parametername> & <value>"
                self.new_checkbox(terms[1],terms[2],terms[3],terms[4])
            elif terms[0]=="RADIOBUTTON":
                assert nterms==5, str_parseerror+"Syntax Error in RADIOBUTTON DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:RADIOBUTTON & <radiobutton_name> & <list_of_options> & <parametername> & <value>"
                self.new_radiobutton(terms[1],terms[2],terms[3],terms[4])
            elif terms[0]=="BUTTON":
                assert nterms==4, str_parseerror+"Syntax Error in BUTTON DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:BUTTON & <button_name> & [<label_text>,<button_text>] & <parametername>"
                self.new_button(terms[1],terms[2],terms[3])
            elif terms[0]=="TEXT_IN":
                assert nterms==6, str_parseerror+"Syntax Error in TEXT_IN DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:TEXT_IN & <textinbox_name> & <size> & <parametername> & <listindex> & <value>"
                self.new_textin(terms[1],terms[2],terms[3],terms[4],terms[5])
            elif terms[0]=="IMAGE":
                assert nterms==6, str_parseerror+"Syntax Error in IMAGE DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:IMAGE & <image_name> & <scale> & <range> & <dataname> & <type>"
                self.new_image(terms[1],terms[2],terms[3],terms[4],terms[5])
            elif terms[0]=="TEXT_OUT":
                assert nterms==5, str_parseerror+"Syntax Error in TEXT_OUT DIRECTIVE, nterms="+str(nterms)+"! "\
                    +"Instead please use syntax @IVISIT:TEXT_OUT & <textoutbox_name> & <size> & <list_of_options> & <dataname>"
                self.new_textout(terms[1],terms[2],terms[3],terms[4])
            else:
                assert 0, "IVisitParser::parse_simulation(.): After @IVISIT: I expected a key word (either SIMULATION, SLIDER, LISTSEL, TEXT_IN, IMAGE, or TEXT_OUT)"\
                    +"at beginning of line "+ln+" in file "+self.filepath+" ! Instead read terms[0]="+terms[0]+"! Note that terms must be separated by '&'!"
            i+=1    # increase main counter of while loop  

    def new_simulation(self,sim_name):
        if not str(self.sim_name).strip() in ['None','NONE','']:   # replace parsed sim_name (given here by sim_name) by constructor sim_name (given here by self.sim_name) ?
            sim_name=self.sim_name
        #print("new_simulation: sim_name=",sim_name)
        now=datetime.now()
        today_str=now. strftime("%Y.%m.%d")
        #print("today_str=",today_str)
        self.simkey=None        # simulation key to be used for following inserts/updates
        if self.db!=None:
            data=self.db.simple_select(['key'],['simulation'],where="name='"+sim_name+"'")
            #print("data=",data)
            if len(data)<=0:    # no simulation with this name --> insert new simulation record in table
                self.simkey=self.db.get_new_primary_key('simulation')[0]
                #print("new key:=",self.simkey)
                self.db.simple_insert('simulation',['key','name','date_init','date_lastmod'],[self.simkey,sim_name,today_str,today_str])   # insert new simulation entry
            else:              # there is already a simulation with this name --> take first, update table
                self.simkey=data[0][0]
                self.db.simple_update_byPKEY('simulation',['date_lastmod'],[today_str],['key'],[self.simkey])
        
    def new_slider(self,sld_name,sld_size,sld_range,sld_parametername,sld_listidx,sld_type,sld_value):
        #print("new_slider: terms=",sld_name,sld_size,sld_range,sld_parametername,sld_listidx,sld_type,sld_value)
        # (i) prepare parameters
        if sld_type in ['int','INT','integer','INTEGER']: sld_type='int'
        if sld_type in ['float','double','FLOAT','DOUBLE']: sld_type='float'
        range_list=parseStringAsList(sld_range,sld_type,[0,100,5,1])
        par_range=str([0,100])
        if len(range_list)>=2: par_range=str(range_list[0:2])
        parwdg_range=str([0,100,5,1])
        if len(range_list)>=4: parwdg_range=str([range_list[0],range_list[1],int(round(range_list[2])),range_list[3]])  # format [min,max,nticks,scale]
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+sld_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range','listidx'],[parkey,sld_parametername,sld_type,par_range,asNumber(sld_listidx,'int',0)]) # insert new parameter entry
            else:                                                                                                                                      # update parameter entry (depending on mode)
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range','listidx'],[sld_type,par_range,asNumber(sld_listidx,'int',0)],['key'],[parkey])   # update all available fields 
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,sld_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name'  ,'type'  ,'range'     ,'size'  ,'value'  ],\
                                      [parwdg_key,self.simkey     ,parkey         ,sld_name,'slider',parwdg_range,sld_size,sld_value])          # insert new parameter entry
            else:                                                                                                                               # update parameter widget entry (depending on mode)
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type', 'range'      ,'size'  ,'value'],\
                                                 ['slider',parwdg_range,sld_size,sld_value], ['key'],[parwdg_key])                              # update all available fields 
                elif self.parsemode in [self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type', 'range'      ,'size'],\
                                                 ['slider',parwdg_range,sld_size], ['key'],[parwdg_key])                                        # update all available fields except value

    def new_dictslider(self,dsld_name,dsld_size,dsld_parametername,dsld_idxselecteditem, dsld_nitems,dsld_itemnames,dsld_ranges,dsld_dictkeys,dsld_types,dsld_values):
        #print("new_dictslider: terms=",dsld_name,dsld_size,dsld_parametername,dsld_idxselecteditem, dsld_nitems,dsld_itemnames,dsld_ranges,dsld_dictkeys,dsld_types,dsld_values)
        # (i) prepare parameters
        par_type='text'                 # parameter value is dict cast as text/string
        par_range = [dsld_nitems]       # use field range in database parameter to store list [nitems, key1,type1,min1,max1, key2,type2,min2,max2, ...] (length 1+nitems*4
        parwdg_range = [dsld_nitems]    # use field range in database parameter to store list [nitems, itemname1,min1,max1,nticks1,scale1, itemname2,min2,max2,nticks2,scale2, ...] (length 1+nitems*5
        parwdg_items = [dsld_nitems,dsld_idxselecteditem]   # information about dict items: [nitems,idxselecteditem,  itemname1,type1,  itemname2,typ2, ...] (length 2+nitems*2) 
        for i in range(dsld_nitems):    # iterate over all items
            # (i.1) handle types
            keyi=dsld_dictkeys[i]       # dict key corresponding to item i
            namei=dsld_itemnames[i].replace(':','') # dict item name corresponding to item i; remove ':' because they cause problems (for display_mode=1)
            typei=dsld_types[i]         # corresponding type of dict value
            if typei in ['int','INT','integer','INTEGER']: typei='int'      # normalize type
            if typei in ['float','double','FLOAT','DOUBLE']: typei='float'  # ditto
            # (i.2) handle ranges
            range_list=parseStringAsList(dsld_ranges[i],typei,[0,100,5,1])
            mini,maxi,nticksi,scalei=0,100,5,1
            if len(range_list)>=2: mini,maxi=range_list[0],range_list[1] 
            if len(range_list)>=4: nticksi,scalei=int(round(range_list[2])),range_list[3]  # format [min,max,nticks,scale]
            # (i.3) add data to attributes
            par_range   +=[keyi,typei,mini,maxi]           # add list of 4 values (key,type,min,max) per item to range (list) of parameter database 
            parwdg_range+=[namei,mini,maxi,nticksi,scalei] # add list of 4 values (name,min,max,nticks,scale) per item
            parwdg_items+=[namei,typei]                    # add list of 2 values (namei,typei) per item
        par_range=str(par_range).replace("'","")       # cast as string to store in database
        parwdg_range=str(parwdg_range).replace("'","") # cast as string to store in database
        parwdg_items=str(parwdg_items).replace("'","") # cast as string to store in database
        parwdg_value='['+", ".join(dsld_values)+']'    # cast as string to store in database
        #print("par_range=",par_range)
        #print("parwdg_range=",parwdg_range)
        #print("parwdg_items=",parwdg_items)
        #print("parwdg_value=",parwdg_value)
        #exit(0)
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+dsld_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range','listidx'],[parkey,dsld_parametername,par_type,par_range,asNumber(dsld_idxselecteditem,'int',0)]) # insert new parameter entry
            else:                                                                                                                                      # update parameter entry (depending on mode)
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range','listidx'],[par_type,par_range,asNumber(dsld_idxselecteditem,'int',0)],['key'],[parkey])   # update all available fields 
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,dsld_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name'   ,'type'      ,'range'     ,'items'     ,'size'   ,'value'  ],\
                                      [parwdg_key,self.simkey     ,parkey         ,dsld_name,'dictslider',parwdg_range,parwdg_items,dsld_size,parwdg_value])          # insert new parameter entry
            else:                                                                                                                               # update parameter widget entry (depending on mode)
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'      ,'range'     ,'items'     ,'size'   ,'value'],\
                                                 ['dictslider',parwdg_range,parwdg_items,dsld_size,parwdg_value], ['key'],[parwdg_key])                       # update all available fields 
                elif self.parsemode in [self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'      ,'range'     ,'items'     ,'size'],\
                                                 ['dictslider',parwdg_range,parwdg_items,dsld_size], ['key'],[parwdg_key])                                    # update all available fields except value

    def new_listsel(self,ls_name,ls_size,ls_options,ls_parametername,ls_listidx,ls_type,ls_value):
        #print("new_listsel: terms=",ls_name,ls_size,ls_options,ls_parametername,ls_listidx,ls_type,ls_value)
        # (i) prepare parameters
        if ls_type in ['string','STRING','text','TEXT']: ls_type='text'
        if ls_type in ['int','INT','integer','INTEGER']: ls_type='int'
        if ls_type in ['float','FLOAT','double','DOUBLE']: ls_type='float'
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+ls_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range','listidx'],[parkey,ls_parametername,ls_type,ls_options,asNumber(ls_listidx,'int',0)])   # insert new parameter entry
            else:
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range','listidx'],[ls_type,ls_options,asNumber(ls_listidx,'int',0)],['key'],[parkey])    # update parameter entry
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,ls_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name' ,'type'         ,'range' ,'size'  ,'value'],\
                                      [parwdg_key,self.simkey     ,parkey         ,ls_name,'listselection',ls_options,ls_size,ls_value])   # insert new parameter entry
            else:
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'         ,'range' ,'size' ,'value'],\
                                                 ['listselection',ls_options,ls_size,ls_value], ['key'],[parwdg_key])   # update parameter widget entry
                elif self.parsemode in [self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'         ,'range' ,'size'],\
                                                 ['listselection',ls_options,ls_size], ['key'],[parwdg_key])            # update parameter widget entry except value
                
    def new_checkbox(self,cb_name,cb_options,cb_parametername,cb_value):
        #print("new_checkbox: terms=",cb_name,cb_options,cb_parametername,cb_value)
        # (i) prepare parameters
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+cb_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range'],[parkey,cb_parametername,'text',cb_options])   # insert new parameter entry
            else:
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range'],['text',cb_options],['key'],[parkey])    # update parameter entry
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,cb_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name' ,'type'    ,'range'   ,'value'],\
                                      [parwdg_key,self.simkey     ,parkey         ,cb_name,'checkbox',cb_options,cb_value])   # insert new parameter entry
            else:
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'    ,'range'   ,'value'],\
                                                 ['checkbox',cb_options,cb_value], ['key'],[parwdg_key])   # update parameter widget entry
                elif self.parsemode in [self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'    ,'range'],\
                                                 ['checkbox',cb_options], ['key'],[parwdg_key])            # update parameter widget entry except value
                
    def new_radiobutton(self,rb_name,rb_options,rb_parametername,rb_value):
        #print("new_radiobutton: terms=",rb_name,rb_options,rb_parametername,rb_value)
        # (i) prepare parameters
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+rb_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range'],[parkey,rb_parametername,'text',rb_options])   # insert new parameter entry
            else:
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range'],['text',rb_options],['key'],[parkey])    # update parameter entry
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,rb_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name' ,'type'       ,'range'   ,'value'],\
                                      [parwdg_key,self.simkey     ,parkey         ,rb_name,'radiobutton',rb_options,rb_value])   # insert new parameter entry
            else:
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'       ,'range'   ,'value'],\
                                                 ['radiobutton',rb_options,rb_value], ['key'],[parwdg_key])   # update parameter widget entry
                elif self.parsemode in [self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'       ,'range'],\
                                                 ['radiobutton',rb_options], ['key'],[parwdg_key])            # update parameter widget entry except value
                
    def new_button(self,rb_name,rb_labeltext_buttontext,rb_parametername):
        #print("new_button: terms=",rb_name,rb_labeltext_buttontext,rb_parametername)
        # (i) prepare parameters
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+rb_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range'],[parkey,rb_parametername,'text',rb_labeltext_buttontext])   # insert new parameter entry
            else:
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range'],['text',rb_labeltext_buttontext],['key'],[parkey])            # update parameter entry
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,rb_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name' ,'type'  ,'range'                ,'value'],\
                                      [parwdg_key,self.simkey     ,parkey         ,rb_name,'button',rb_labeltext_buttontext,'0'])      # insert new parameter entry
            else:
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'  ,'range'                ,'value'],\
                                                 ['button',rb_labeltext_buttontext,'0'    ], ['key'],[parwdg_key])   # update parameter widget entry (value is here irrelevant as set to '0')
                
    def new_textin(self,txi_name,txi_size,txi_parametername,txi_listidx,txi_value):
        #print("new_textin: terms=",txi_name,txi_size,txi_parametername,txi_listidx,txi_value)
        # (i) prepare parameters
        if self.db!=None:
            # (ii) Create/Modify parameter entry in table 'parameter'
            parkey=None                            # find parameter key
            data=self.db.simple_select(['key'],['parameter'],where="name='"+txi_parametername+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                parkey = self.db.get_new_primary_key('parameter')[0]
                self.db.simple_insert('parameter',['key','name','type','range','listidx'],[parkey,txi_parametername,'text','None',asNumber(txi_listidx,'int',0)])   # insert new parameter entry
            else:
                parkey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameter',['type','range','listidx'],['text','None',asNumber(txi_listidx,'int',0)],['key'],[parkey])    # update parameter entry
            # (iii) Create/Modify parameter widget entry in table 'parameterwidget'
            parwdg_key=None                        # find parameter key
            whereclause=self.db.getWhereClause_from_ColumnValues(['parameterwidget'],['key_simulation','key_parameter','name'],[self.simkey,parkey,txi_name])
            data=self.db.simple_select(['key'],['parameterwidget'],where=whereclause)
            if len(data)<=0: # no parameter widget available --> insert new one in table 
                parwdg_key = self.db.get_new_primary_key('parameterwidget')[0]
                self.db.simple_insert('parameterwidget',\
                                      ['key'     ,'key_simulation','key_parameter','name'  ,'type'     ,'range' ,'size'  ,'value'],\
                                      [parwdg_key,self.simkey     ,parkey         ,txi_name,'textfield','None'  ,txi_size,txi_value])   # insert new parameter entry
            else:
                parwdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'     ,'range','size'  ,'value'  ],\
                                                 ['textfield','None' ,txi_size,txi_value], ['key'],[parwdg_key])   # update parameter widget entry
                elif self.parsemode in [self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('parameterwidget',\
                                                 ['type'     ,'range','size'],\
                                                 ['textfield','None' ,txi_size], ['key'],[parwdg_key])             # update parameter widget entry except value
                
    def new_image(self,img_name,img_scale,img_range,img_dataname,img_type):
        #print("new_image: terms=",img_name,img_scale,img_range,img_dataname,img_type)
        # (i) prepare parameters
        if img_type in ['int','INT','integer','INTEGER']: img_type='int'
        if img_type in ['float','FLOAT','double','DOUBLE']: img_type='float'
        if img_type in ['binary','bool','boolean','BINARY','BOOL','BOOLEAN']: img_type='binary'
        if self.db!=None:
            # (ii) Create/Modify dataarray entry in table 'dataarray'
            datakey=None                            # find dataarray key
            data=self.db.simple_select(['key'],['dataarray'],where="name='"+img_dataname+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                datakey = self.db.get_new_primary_key('dataarray')[0]
                self.db.simple_insert('dataarray',['key','name','type','range'],[datakey,img_dataname,img_type,img_range])  # insert new dataarray entry
            else:
                datakey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('dataarray',['type','range'],[img_type,img_range],['key'],[datakey])           # update dataarray entry
            # (iii) Create/Modify data widget entry in table 'datawidget'
            datawdg_key=None                        # find data widget key
            whereclause=self.db.getWhereClause_from_ColumnValues(['datawidget'],['key_simulation','key_dataarray','name'],[self.simkey,datakey,img_name])
            data=self.db.simple_select(['key'],['datawidget'],where=whereclause)
            if len(data)<=0: # no data widget available --> insert new one in table 
                datawdg_key = self.db.get_new_primary_key('datawidget')[0]
                self.db.simple_insert('datawidget',\
                                      ['key'      ,'key_simulation','key_dataarray','name'  ,'type'  ,'range'  ,'scale'],\
                                      [datawdg_key,self.simkey     ,datakey        ,img_name,'image' ,img_range,img_scale])   # insert new parameter entry
            else:
                datawdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('datawidget',\
                                                 ['type' ,'range'  ,'scale'  ],\
                                                 ['image',img_range,img_scale], ['key'],[datawdg_key])                        # update parameter widget entry
                
    def new_textout(self,txo_name,txo_size,txo_options,txo_dataname):
        #print("new_textout: terms=",txo_name,txo_size,txo_dataname)
        # (i) prepare parameters
        options_list=parseStringAsList(txo_options,'string',['None'])    # just for checking and normalizing
        txo_options=str(options_list)
        txo_options=txo_options.replace("'","")                          # remove quotes
        txo_options=txo_options.replace('"','')                          # remove quotes 
        if self.db!=None:
            # (ii) Create/Modify dataarray entry in table 'dataarray'
            datakey=None                            # find dataarray key
            data=self.db.simple_select(['key'],['dataarray'],where="name='"+txo_dataname+"'")
            if len(data)<=0: # no parameter with this name --> insert new one in table
                datakey = self.db.get_new_primary_key('dataarray')[0]
                self.db.simple_insert('dataarray',['key','name','type','range'],[datakey,txo_dataname,'text','None'])   # insert new dataarray entry
            else:
                datakey=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('dataarray',['type','range'],['text','None'],['key'],[datakey])        # update dataarray entry
            # (iii) Create/Modify data widget entry in table 'datawidget'
            datawdg_key=None                        # find data widget key
            whereclause=self.db.getWhereClause_from_ColumnValues(['datawidget'],['key_simulation','key_dataarray','name'],[self.simkey,datakey,txo_name])
            data=self.db.simple_select(['key'],['datawidget'],where=whereclause)
            if len(data)<=0: # no data widget available --> insert new one in table 
                datawdg_key = self.db.get_new_primary_key('datawidget')[0]
                self.db.simple_insert('datawidget',\
                                      ['key'      ,'key_simulation','key_dataarray','name'  ,'type'     ,'range'    ,'scale'],\
                                      [datawdg_key,self.simkey     ,datakey        ,txo_name,'textfield',txo_options,1.0])   # insert new data widget entry
            else:
                datawdg_key=data[0][0]
                if self.parsemode in [self.PARSEMODE_UpdateOld,self.PARSEMODE_UpdateOldExceptValues]:
                    self.db.simple_update_byPKEY('datawidget',\
                                                 ['type'     ,'range'    ,'scale'],\
                                                 ['textfield',txo_options,1.0    ], ['key'],[datawdg_key])   # update data widget entry
                
if __name__ == '__main__':
    print("\nModule Test ivisit.parser (A.Knoblauch, 15/9/2019)") 
    print("---------------------------------------------------\n") 
    pr = IVisitParser("dummyscripttoparse.py") 

